from abc import ABC, abstractmethod
from collections import namedtuple
from typing import Dict, Union

from .apdu import ApduServer
from .seproxyhal import SeProxyHal
from .button_tcp import FakeButton
from .finger_tcp import FakeFinger
from .vnc import VNC

Server = Union[ApduServer, FakeButton, FakeFinger, SeProxyHal, VNC]

DisplayArgs = namedtuple("DisplayArgs", "color model ontop rendering keymap pixel_size x y")
ServerArgs = namedtuple("ServerArgs", "apdu apirun button finger seph vnc")

Model = namedtuple('Model', 'name screen_size box_position box_size')
MODELS = {
    'nanos': Model('Nano S', (128, 32), (20, 13), (100, 26)),
    'nanox': Model('Nano X', (128, 64), (5, 5), (10, 10)),
    'blue': Model('Blue', (320, 480), (13, 13), (26, 26)),
}

COLORS = {
    'LAGOON_BLUE': 0x7ebab5,
    'JADE_GREEN': 0xb9ceac,
    'FLAMINGO_PINK': 0xd8a0a6,
    'SAFFRON_YELLOW': 0xf6a950,
    'MATTE_BLACK': 0x111111,

    'CHARLOTTE_PINK': 0xff5555,
    'ARNAUD_GREEN': 0x79ff79,
    'SYLVE_CYAN': 0x29f3f3,
}


class Screenshot:
    def __init__(self, screen_size):
        self.pixels = {}
        self.width, self.height = screen_size
        for y in range(0, self.height):
            for x in range(0, self.width):
                self.pixels[(x, y)] = 0x000000

    def update(self, pixels):
        # Don't call update, replace the object instead
        self.pixels = {**self.pixels, **pixels}

    def get_image(self):
        # Get the pixels object once, as it may be replaced during the loop.
        pixels = self.pixels
        data = bytearray(self.width * self.height * 3)
        for y in range(0, self.height):
            for x in range(0, self.width):
                pos = 3 * (y * self.width + x)
                data[pos:pos + 3] = pixels[(x, y)].to_bytes(3, "big")
        return (self.width, self.height), bytes(data)


class FrameBuffer(ABC):
    COLORS = {
        "nanos": 0x00fffb,
        "nanox": 0xdddddd,
    }

    def __init__(self, model):
        self.pixels = {}
        self.model = model
        self.screenshot = Screenshot(MODELS[model].screen_size)

    def draw_point(self, x, y, color):
        # There are only 2 colors on the Nano S and the Nano X but the one
        # passed in argument isn't always valid. Fix it here.
        if color != 0x000000:
            color = FrameBuffer.COLORS.get(self.model, color)
        self.pixels[(x, y)] = color

    def screenshot_update_pixels(self):
        self.screenshot.update(self.pixels)

    def take_screenshot(self):
        return self.screenshot.get_image()


class Display(ABC):
    def __init__(self, display: DisplayArgs, server: ServerArgs) -> None:
        self.notifiers: Dict[int, Server] = {}
        self.apdu = server.apdu
        self.seph = server.seph
        self.model = display.model
        self.rendering = display.rendering

    @abstractmethod
    def display_status(self, data):
        pass

    @abstractmethod
    def display_raw_status(self, data):
        pass

    @abstractmethod
    def screen_update(self) -> bool:
        pass

    def add_notifier(self, klass):
        assert klass.s.fileno() not in self.notifiers
        self.notifiers[klass.s.fileno()] = klass

    def remove_notifier(self, fd):
        self.notifiers.pop(fd)

    def _init_notifiers(self, args: ServerArgs) -> None:
        for klass in args._asdict().values():
            if klass:
                self.add_notifier(klass)

    def forward_to_app(self, packet):
        self.seph.to_app(packet)

    def forward_to_apdu_client(self, packet):
        self.apdu.forward_to_client(packet)



import binascii
import logging
from collections import namedtuple
from construct import Aligned, Struct, Int8ul, Int16ul, Int32ul, Padded

from . import bagl_font
from . import bagl_glyph
from .automation import TextEvent

bagl_component_t = Aligned(4, Struct(
    "type" / Int8ul,
    "userid" / Int8ul,
    "x" / Int16ul,
    "y" / Int16ul,
    "width" / Int16ul,
    "height" / Int16ul,
    "stroke" / Int8ul,
    "radius" / Int8ul,
    "fill" / Padded(4, Int8ul),
    "fgcolor" / Int32ul,
    "bgcolor" / Int32ul,
    "font_id" / Int16ul,
    "icon_id" / Int8ul,
))

BAGL_FILL = 1

BAGL_FILL_CIRCLE_1_OCTANT = 1
BAGL_FILL_CIRCLE_2_OCTANT = 2
BAGL_FILL_CIRCLE_3_OCTANT = 4
BAGL_FILL_CIRCLE_4_OCTANT = 8
BAGL_FILL_CIRCLE_5_OCTANT = 16
BAGL_FILL_CIRCLE_6_OCTANT = 32
BAGL_FILL_CIRCLE_7_OCTANT = 64
BAGL_FILL_CIRCLE_8_OCTANT = 128
BAGL_FILL_CIRCLE = 0xFF
BAGL_FILL_CIRCLE_3PI2_2PI = (BAGL_FILL_CIRCLE_1_OCTANT | BAGL_FILL_CIRCLE_2_OCTANT)
BAGL_FILL_CIRCLE_PI_3PI2 = (BAGL_FILL_CIRCLE_3_OCTANT | BAGL_FILL_CIRCLE_4_OCTANT)
BAGL_FILL_CIRCLE_0_PI2 = (BAGL_FILL_CIRCLE_5_OCTANT | BAGL_FILL_CIRCLE_6_OCTANT)
BAGL_FILL_CIRCLE_PI2_PI = (BAGL_FILL_CIRCLE_7_OCTANT | BAGL_FILL_CIRCLE_8_OCTANT)

BAGL_NONE = 0
BAGL_BUTTON = 1
BAGL_LABEL = 2
BAGL_RECTANGLE = 3
BAGL_LINE = 4
BAGL_ICON = 5
BAGL_CIRCLE = 6
BAGL_LABELINE = 7
BAGL_FLAG_TOUCHABLE = 0x80
BAGL_TYPE_FLAGS_MASK = 0x80

BAGL_FONT_ALIGNMENT_HORIZONTAL_MASK = 0xC000
BAGL_FONT_ALIGNMENT_LEFT = 0x0000
BAGL_FONT_ALIGNMENT_RIGHT = 0x4000
BAGL_FONT_ALIGNMENT_CENTER = 0x8000
BAGL_FONT_ALIGNMENT_VERTICAL_MASK = 0x3000
BAGL_FONT_ALIGNMENT_TOP = 0x0000
BAGL_FONT_ALIGNMENT_BOTTOM = 0x1000
BAGL_FONT_ALIGNMENT_MIDDLE = 0x2000

SEPROXYHAL_TAG_SCREEN_DISPLAY_RAW_STATUS_START = 0x00

DrawState = namedtuple('DrawState', 'x y width height colors bpp xx yy')


class Bagl:
    def __init__(self, m, size):
        self.m = m
        self.SCREEN_WIDTH, self.SCREEN_HEIGHT = size
        self.draw_state = DrawState(0, 0, 0, 0, [], 0, 0, 0)
        self.logger = logging.getLogger("bagl")

    def refresh(self) -> bool:
        return self.m.update()

    def hal_draw_bitmap_within_rect(self, x, y, width, height, colors, bpp, bitmap, restore=None):
        if bpp == 3 or bpp > 4:
            return

        if x >= self.SCREEN_WIDTH or y >= self.SCREEN_HEIGHT:
            return

        if x + width > self.SCREEN_WIDTH:
            width = self.SCREEN_WIDTH - x

        if y + height > self.SCREEN_HEIGHT:
            height = self.SCREEN_HEIGHT - y

        pixel_mask = (1 << bpp) - 1

        requested_xw = width + x
        requested_yh = height + y

        if not restore:
            xx, yy = x, y
        else:
            xx, yy = restore

        bitmap = list(bitmap)
        while bitmap:
            ch = bitmap.pop(0)
            for i in range(0, 8, bpp):
                if xx >= 0 and xx < self.SCREEN_WIDTH and yy >= 0 and yy < self.SCREEN_HEIGHT:
                    pixel_color_index = (ch >> i) & pixel_mask
                    color = colors[pixel_color_index]
                    self.m.draw_point(xx, yy, color)

                xx += 1
                if xx >= requested_xw:
                    xx = x
                    yy += 1
                    if yy >= requested_yh:
                        self.draw_state = DrawState(x, y, width, height, colors, bpp, xx, yy)
                        return

        self.draw_state = DrawState(x, y, width, height, colors, bpp, xx, yy)

    def hal_draw_rect(self, color, x, y, width, height):
        if x + width > self.SCREEN_WIDTH or x < 0:
            if x > self.SCREEN_WIDTH:
                return
            else:
                width = self.SCREEN_WIDTH - x
        if y + height > self.SCREEN_HEIGHT or y < 0:
            if y > self.SCREEN_HEIGHT:
                return
            else:
                height = self.SCREEN_HEIGHT - y

        YX = (y//8)*self.SCREEN_WIDTH + x
        YXlinemax = YX + width
        xx = x
        requested_x = x

        i = width * height
        while i > 0:
            i -= 1
            self.m.draw_point(xx, y, color)
            YX += 1
            xx += 1
            if YX >= YXlinemax:
                y += 1
                height -= 1
                YX = (y // 8) * self.SCREEN_WIDTH + x
                YXlinemax = YX + width
                xx = requested_x
            if height == 0:
                break

    @staticmethod
    def compute_line_width(font_id, width, text, text_encoding):
        font = bagl_font.get(font_id)
        if not font:
            logging.error("font not found")
            return 0

        xx = 0

        text = text.replace(b'\r\n', b'\n')
        for ch in text:
            ch_width = 0

            if ch < font.first_char or ch > font.last_char:
                if ch in ['\r', '\n']:
                    return xx

                if ch >= 0xc0:
                    ch_width = ch & 0x3f
                elif ch >= 0x80:
                    if ch & 0x20:
                        font_symbol_id = bagl_font.BAGL_FONT_SYMBOLS_1
                    else:
                        font_symbol_id = bagl_font.BAGL_FONT_SYMBOLS_0
                    font_symbols = bagl_font.get(font_symbol_id)
                    if font_symbols:
                        ch_width = font.characters[ch & 0x1f].char_width
            else:
                ch -= font.first_char
                ch_width = font.characters[ch].char_width

            if xx + ch_width > width and width > 0:
                return xx

            xx += ch_width

        return xx

    def draw_string(self, font_id, fgcolor, bgcolor, x, y, width, height, text, text_encoding):
        font = bagl_font.get(font_id)
        if not font:
            self.logger.error("unsupported font {}".format(font_id & bagl_font.BAGL_FONT_ID_MASK))
            return 0

        if font.bpp > 1:
            color_count = (1 << font.bpp) - 1
            colors = [bgcolor]
            for i in range(1, color_count):
                colors.append(0)

            for off in range(0, 3):
                cfg = (fgcolor >> (off * 8)) & 0xFF
                cbg = (bgcolor >> (off * 8)) & 0xFF

                crange = max(cfg, cbg) - min(cfg, cbg) + 1
                cinc = crange // color_count

                if cfg > cbg:
                    for i in range(1, color_count):
                        colors[i] |= min(0xff, cbg + i * cinc) << (off * 8)
                else:
                    for i in range(1, color_count):
                        colors[i] |= min(0xff, cfg + (color_count - i) * cinc) << (off * 8)

            colors.append(fgcolor)
        else:
            colors = [bgcolor, fgcolor]

        width += x
        height += y
        xx = x

        text = text.replace(b'\r\n', b'\n')

        for ch in text:
            ch_height = font.char_height
            ch_kerning = 0
            ch_width = 0
            ch_bitmap = None
            ch_y = y

            if ch < font.first_char or ch > font.last_char:
                if ch in ['\r', '\n']:
                    y += ch_height
                    if y + ch_height > height:
                        return (y << 16) | (xx & 0xFFFF)
                    xx = x
                    continue

                if ch >= 0xc0:
                    ch_width = ch & 0x3f
                elif ch >= 0x80:
                    if ch & 0x20:
                        font_symbol_id = bagl_font.BAGL_FONT_SYMBOLS_1
                    else:
                        font_symbol_id = bagl_font.BAGL_FONT_SYMBOLS_0
                    font_symbols = bagl_font.get(font_symbol_id)
                    if font_symbols:
                        ch_bitmap = font.bitmap[font.characters[ch & 0x1f].bitmap_offset:]
                        ch_width = font.characters[ch & 0x1f].char_width
                        ch_height = font_symbols.char_height
                        ch_y = y + font.baseline_height - font_symbols.baseline_height
            else:
                ch -= font.first_char
                ch_bitmap = font.bitmap[font.characters[ch].bitmap_offset:]
                ch_width = font.characters[ch].char_width
                ch_kerning = font.char_kerning

            if xx + ch_width > width:
                y += ch_height
                if y + ch_height > height:
                    return (y << 16) | (xx & 0xFFFF)
                xx = x
                ch_y = y

            if ch_bitmap:
                self.hal_draw_bitmap_within_rect(xx, ch_y, ch_width, ch_height, colors, font.bpp, ch_bitmap)
            else:
                self.hal_draw_rect(bgcolor, xx, ch_y, ch_width, ch_height)

            xx += ch_width + ch_kerning

        return (y << 16) | (xx & 0xFFFF)

    def _draw_circle_helper(self, color, x_center, y_center, radius, octants, radiusint, colorint):
        x, y = radius, 0
        decisionOver2 = 1 - x
        dradius = radius-radiusint
        last_x = x
        drawint = (radiusint > 0 and dradius > 0)

        while y <= x:
            if octants & 1:
                if drawint:
                    self.hal_draw_rect(colorint, x_center,   y+y_center, x-(dradius-1), 1)
                    self.hal_draw_rect(color, x_center+x-(dradius-1), y+y_center, dradius, 1)
                else:
                    self.hal_draw_rect(color, x_center,   y+y_center-1, x, 1)
            if octants & 2:
                if drawint:
                    if last_x != x:
                        self.hal_draw_rect(colorint, x_center,   x+y_center, y-(dradius-1), 1)
                    self.hal_draw_rect(color, x_center+y-(dradius-1), x+y_center, dradius, 1)
                else:
                    self.hal_draw_rect(color, x_center,   x+y_center-1, y, 1)
            if octants & 4:
                if drawint:
                    self.hal_draw_rect(colorint, x_center-x, y+y_center, x-(dradius-1), 1)
                    self.hal_draw_rect(color, x_center-x-(dradius-1), y+y_center, dradius, 1)
                else:
                    self.hal_draw_rect(color, x_center-x, y+y_center-1, x, 1)
            if octants & 8:
                if drawint:
                    if last_x != x:
                        self.hal_draw_rect(colorint, x_center-y, x+y_center, y-(dradius-1), 1)
                    self.hal_draw_rect(color, x_center-y-(dradius-1), x+y_center, dradius, 1)
                else:
                    self.hal_draw_rect(color, x_center-y, x+y_center-1, y, 1)
            if octants & 16:
                if drawint:
                    self.hal_draw_rect(colorint, x_center,   y_center-y, x-(dradius-1), 1)
                    self.hal_draw_rect(color, x_center+x-(dradius-1), y_center-y, dradius, 1)
                else:
                    self.hal_draw_rect(color, x_center,   y_center-y, x, 1)
            if octants & 32:
                if drawint:
                    if last_x != x:
                        self.hal_draw_rect(colorint, x_center,   y_center-x, y-(dradius-1), 1)
                    self.hal_draw_rect(color, x_center+y-(dradius-1), y_center-x, dradius, 1)
                else:
                    self.hal_draw_rect(color, x_center,   y_center-x, y, 1)
            if octants & 64:
                if drawint:
                    self.hal_draw_rect(colorint, x_center-x, y_center-y, x-(dradius-1), 1)
                    self.hal_draw_rect(color, x_center-x-(dradius-1), y_center-y, dradius, 1)
                else:
                    self.hal_draw_rect(color, x_center-x, y_center-y, x, 1)
            if octants & 128:
                if drawint:
                    if last_x != x:
                        self.hal_draw_rect(colorint, x_center-y, y_center-x, y-(dradius-1), 1)
                    self.hal_draw_rect(color, x_center-y-(dradius-1), y_center-x, dradius, 1)
                else:
                    self.hal_draw_rect(color, x_center-y, y_center-x, y, 1)

            last_x = x
            y += 1
            if decisionOver2 <= 0:
                decisionOver2 += 2 * y + 1
            else:
                x -= 1
                decisionOver2 += 2 * (y - x) + 1

    def _display_bagl_icon(self, component, context):
        if component.icon_id != 0:
            self.logger.debug(f"icon_id {component.icon_id}")
            glyph = bagl_glyph.get(component.icon_id)
            if not glyph:
                self.logger.error(f"glyph {component.icon_id:#x} not found")
                return

            if len(context) != 0:
                # assert (1 << glyph.bpp) * 4 == len(context)
                colors = []
                for i in range(0, 1 << glyph.bpp):
                    color = int.from_bytes(context[i*4:(i+1)*4], byteorder='big')
                    colors.append(color)
            else:
                colors = glyph.colors

            self.hal_draw_bitmap_within_rect(
                component.x + (component.width // 2 - glyph.width // 2),
                component.y + (component.height // 2 - glyph.height // 2),
                glyph.width,
                glyph.height,
                colors,
                glyph.bpp,
                glyph.bitmap)
        else:
            if len(context) == 0:
                self.logger.info("len context == 0 {}".format(binascii.hexlify(context)))
                return

            bpp = context[0]
            if bpp > 2:
                return
            colors = []
            n = 1
            for i in range(0, 1 << bpp):
                color = int.from_bytes(context[n:n+4], byteorder='big')
                colors.append(color)
                n += 4
            bitmap = context[n:]
            bitmap_length_bits = bpp * component.width * component.height
            assert len(bitmap) * 8 >= bitmap_length_bits
            self.hal_draw_bitmap_within_rect(component.x, component.y,
                                             component.width, component.height,
                                             colors,
                                             bpp,
                                             bitmap)

    def _display_bagl_rectangle(self, component, context, context_encoding, halignment, valignment):
        radius = min(component.radius, min(component.width//2, component.height//2))
        if component.fill != BAGL_FILL:
            coords = [
                # centered top to bottom
                (component.x+radius, component.y, component.width-2*radius, component.height),
                # left to center rect
                (component.x, component.y+radius, radius, component.height-2*radius),
                # center rect to right
                (component.x+component.width-radius-1, component.y+radius, radius, component.height-2*radius),
            ]
            for (x, y, width, height) in coords:
                self.hal_draw_rect(component.bgcolor, x, y, width, height)
            coords = [
                # outline. 4 rectangles (with last pixel of each corner not set)
                # top, bottom, left, right
                (component.x+radius, component.y, component.width-2*radius, component.stroke),
                (component.x+radius, component.y+component.height-1, component.width-2*radius, component.stroke),
                (component.x, component.y+radius, component.stroke, component.height-2*radius),
                (component.x+component.width-1, component.y+radius, component.stroke, component.height-2*radius),
            ]
            for (x, y, width, height) in coords:
                self.hal_draw_rect(component.fgcolor, x, y, width, height)
        else:
            coords = [
                # centered top to bottom
                (component.x+radius, component.y, component.width-2*radius, component.height),
                # left to center rect
                (component.x, component.y+radius, radius, component.height-2*radius),
                # center rect to right
                (component.x+component.width-radius, component.y+radius, radius, component.height-2*radius),
            ]
            for (x, y, width, height) in coords:
                self.hal_draw_rect(component.fgcolor, x, y, width, height)

        if radius > 1:
            radiusint = 0
            if component.fill != BAGL_FILL and component.stroke < radius:
                radiusint = radius - component.stroke
            self._draw_circle_helper(component.fgcolor,
                                     component.x + radius,
                                     component.y + radius,
                                     radius,
                                     BAGL_FILL_CIRCLE_PI2_PI,
                                     radiusint,
                                     component.bgcolor)
            self._draw_circle_helper(component.fgcolor,
                                     component.x + component.width - radius - component.stroke,
                                     component.y + radius,
                                     radius,
                                     BAGL_FILL_CIRCLE_0_PI2,
                                     radiusint,
                                     component.bgcolor)
            self._draw_circle_helper(component.fgcolor,
                                     component.x + radius,
                                     component.y + component.height - radius - component.stroke,
                                     radius,
                                     BAGL_FILL_CIRCLE_PI_3PI2,
                                     radiusint,
                                     component.bgcolor)
            self._draw_circle_helper(component.fgcolor,
                                     component.x + component.width - radius - component.stroke,
                                     component.y + component.height - radius - component.stroke,
                                     radius,
                                     BAGL_FILL_CIRCLE_3PI2_2PI,
                                     radiusint,
                                     component.bgcolor)

        if context:
            fgcolor, bgcolor = component.fgcolor, component.bgcolor
            if component.fill == BAGL_FILL:
                fgcolor, bgcolor = bgcolor, fgcolor
            stroke = max(1, component.stroke * 2)
            self.draw_string(component.font_id,
                             fgcolor,
                             bgcolor,
                             component.x + halignment,	         	 # XXX: take icon_width into account
                             component.y + valignment,
                             component.width - halignment - stroke,	 # XXX: take icon_width into account
                             component.height - valignment - stroke,
                             context,
                             context_encoding)

    def _display_bagl_labeline(self, component, text, halignment, valignment, baseline, char_height, strwidth, type_):
        if component.fill == BAGL_FILL:
            y = component.y
            height = component.height
            if type_ == BAGL_LABELINE:
                y -= baseline
                height = char_height
            self.hal_draw_rect(component.bgcolor, component.x, y, halignment, height)
            self.hal_draw_rect(component.bgcolor, component.x + halignment + strwidth,
                               y, component.width - (halignment + strwidth), height)

        if len(text) == 0:
            return

        # XXX
        context_encoding = 0
        y = component.y
        height = component.height
        if type_ == BAGL_LABELINE:
            y -= baseline
        else:
            y += valignment
            height += valignment
        self.draw_string(component.font_id,
                         component.fgcolor,
                         component.bgcolor,
                         component.x + halignment,
                         y,
                         component.width - halignment,
                         component.height,
                         text,
                         context_encoding)

        return [TextEvent(text.decode("utf-8", "ignore"), component.x + halignment, y)]

    def _display_get_alignment(self, component, context, context_encoding):
        halignment = 0
        valignment = 0
        baseline = 0
        char_height = 0
        strwidth = 0

        if component.type == BAGL_ICON:
            return (halignment, valignment, baseline, char_height, strwidth)

        font = bagl_font.get(component.font_id)
        if font:
            baseline = font.baseline_height
            char_height = font.char_height

            if context:
                strwidth = Bagl.compute_line_width(component.font_id,
                                                   component.width + 100,
                                                   context,
                                                   context_encoding)

        haligned = (component.font_id & BAGL_FONT_ALIGNMENT_HORIZONTAL_MASK)
        if haligned == BAGL_FONT_ALIGNMENT_RIGHT:
            halignment = component.width - strwidth
        elif haligned == BAGL_FONT_ALIGNMENT_CENTER:
            halignment = component.width // 2 - strwidth // 2
        elif haligned == BAGL_FONT_ALIGNMENT_LEFT:
            halignment = 0
        else:
            halignment = 0

        valigned = (component.font_id & BAGL_FONT_ALIGNMENT_VERTICAL_MASK)
        if valigned == BAGL_FONT_ALIGNMENT_BOTTOM:
            valignment = component.height - baseline
        elif valigned == BAGL_FONT_ALIGNMENT_MIDDLE:
            valignment = component.height // 2 - baseline // 2 - 1
        elif valigned == BAGL_FONT_ALIGNMENT_TOP:
            valignment = 0
        else:
            valignment = 0

        return (halignment, valignment, baseline, char_height, strwidth)

    def display_status(self, data):
        component = bagl_component_t.parse(data)
        context = data[bagl_component_t.sizeof():]
        context_encoding = 0  # XXX
        self.logger.debug("component: {}".format(component))

        ret = self._display_get_alignment(component, context, context_encoding)
        (halignment, valignment, baseline, char_height, strwidth) = ret

        ret = None
        type_ = component.type & (~BAGL_TYPE_FLAGS_MASK)
        if type_ == BAGL_NONE:
            # TODO
            # self.renderer.clear()
            pass
        elif type_ == BAGL_RECTANGLE:
            self._display_bagl_rectangle(component, context, context_encoding, halignment, valignment)
        elif type_ == BAGL_LABEL:
            self._display_bagl_labeline(component, context, halignment, valignment, baseline, char_height,
                                        strwidth, type_)
        elif type_ == BAGL_LABELINE:
            ret = self._display_bagl_labeline(component, context, halignment, valignment, baseline, char_height,
                                              strwidth, type_)
        elif type_ == BAGL_ICON:
            self._display_bagl_icon(component, context)
        return ret

    def display_raw_status(self, data):
        if data[0] == SEPROXYHAL_TAG_SCREEN_DISPLAY_RAW_STATUS_START:
            x = int.from_bytes(data[1:3], byteorder='big', signed=True)
            y = int.from_bytes(data[3:5], byteorder='big', signed=True)
            w = int.from_bytes(data[5:7], byteorder='big')
            h = int.from_bytes(data[7:9], byteorder='big')
            bpp = int.from_bytes(data[9:10], byteorder='big')

            color_size = 4 * (1 << bpp)
            colors = []
            for i in range(10, 10 + color_size, 4):
                color = int.from_bytes(data[i:i+4], byteorder='little')
                colors.append(color)
            bitmap = data[10+color_size:]

            self.hal_draw_bitmap_within_rect(x, y, w, h, colors, bpp, bitmap)

        else:
            bitmap = data[1:]

            x = self.draw_state.x
            y = self.draw_state.y
            w = self.draw_state.width
            h = self.draw_state.height
            colors = self.draw_state.colors
            bpp = self.draw_state.bpp
            restore = (self.draw_state.xx, self.draw_state.yy)

            self.hal_draw_bitmap_within_rect(x, y, w, h, colors, bpp, bitmap, restore)



import sys
import select
import curses
import logging
import os

from . import bagl
from .display import Display, DisplayArgs, FrameBuffer, MODELS, ServerArgs
from .readerror import ReadError
import time

wait_time = 0.01

BUTTON_LEFT = 1
BUTTON_RIGHT = 2

_TEXT_ = "\033[36;40m"
_BORDER_ = "\033[30;1;40m"
_RESET_COLOR = "\033[0m"

M = [0]*16
M[0b0000] = ' '
M[0b0001] = '\u2598'
M[0b0010] = '\u259D'
M[0b0011] = '\u2580'
M[0b0100] = '\u2596'
M[0b0101] = '\u258C'
M[0b0110] = '\u259E'
M[0b0111] = '\u259B'
M[0b1000] = '\u2597'
M[0b1001] = '\u259A'
M[0b1010] = '\u2590'
M[0b1011] = '\u259C'
M[0b1100] = '\u2584'
M[0b1101] = '\u2599'
M[0b1110] = '\u259F'
M[0b1111] = '\u2588'


# a b
# c d
def map_pix(a, b, c, d):
    return M[d << 3 | c << 2 | b << 1 | a]


class TextWidget(FrameBuffer):
    def __init__(self, parent, model):
        super().__init__(model)
        self.width = parent.width
        self.height = parent.height
        self.previous_screen = 0

        # ncurses stops the process if in the background
        if os.tcgetpgrp(sys.stdin.fileno()) != os.getpgrp():
            logging.getLogger("display").warn("please run speculos in the foreground to allow the initialization of "
                                              "the display")

        self.stdscr = curses.initscr()
        curses.noecho()
        curses.cbreak()
        curses.start_color()
        curses.init_pair(1, curses.COLOR_CYAN, curses.COLOR_BLACK)
        curses.init_pair(2, curses.COLOR_BLACK, curses.COLOR_WHITE)
        self.stdscr.nodelay(1)    # returns -1 if nothing
        self.stdscr.clear()
        self.stdscr.keypad(1)   # interpret escape sequences generated by keypad and function keys

    def get_pixel(self, x, y):
        color = self.pixels.get((x, y), 0)
        return int(color != 0)

    def _redraw(self):
        if self.pixels == self.previous_screen:
            return False

        self.previous_screen = self.pixels.copy()

        self.stdscr.clear()
        for i in range(0, self.height-2, 2):
            line = []
            for j in range(0, self.width-2, 2):
                a = self.get_pixel(j, i)
                b = self.get_pixel(j+1, i)
                c = self.get_pixel(j, i+1)
                d = self.get_pixel(j+1, i+1)
                line.append(map_pix(a, b, c, d))

            self.stdscr.addstr(1 + i//2, 0, ' ', curses.color_pair(2))
            self.stdscr.addstr(1 + i//2, 1, ''.join(line), curses.color_pair(1))
            self.stdscr.addstr(1 + i//2, self.width//2 + 1, ' ', curses.color_pair(2))

        self.stdscr.addstr(0, 0, ' '*(self.width//2 + 2), curses.color_pair(2))
        self.stdscr.addstr(self.height//2, 0, ' '*(self.width//2 + 2), curses.color_pair(2))
        self.stdscr.refresh()

    def draw_point(self, x, y, color):
        self.pixels[(x, y)] = color
        self.screenshot_update_pixels()


class TextScreen(Display):
    def __init__(self, display: DisplayArgs, server: ServerArgs) -> None:
        super().__init__(display, server)

        self.width, self.height = MODELS[display.model].screen_size
        self.m = TextWidget(self, display.model)
        self.bagl = bagl.Bagl(self.m, MODELS[display.model].screen_size)

        self._init_notifiers(server)

        if display.keymap is not None:
            self.ARROW_KEYS = list(map(ord, display.keymap))
        else:
            self.ARROW_KEYS = [curses.KEY_LEFT, curses.KEY_RIGHT, curses.KEY_DOWN]

        self.key2btn = {
                            self.ARROW_KEYS[0]: BUTTON_LEFT,
                            self.ARROW_KEYS[1]: BUTTON_RIGHT,
                            self.ARROW_KEYS[2]: BUTTON_LEFT | BUTTON_RIGHT,
                        }

    def display_status(self, data):
        return self.bagl.display_status(data)

    def display_raw_status(self, data):
        self.bagl.display_raw_status(data)

    def screen_update(self) -> bool:
        return self.m._redraw()

    def get_keypress(self):
        key = self.m.stdscr.getch()
        if key == -1:
            return True
        elif key in self.ARROW_KEYS:
            self.seph.handle_button(self.key2btn[key], 1)
            time.sleep(wait_time)
            self.seph.handle_button(self.key2btn[key], 0)
            return True
        elif key == ord('q'):
            return False
        else:
            return True

    def run(self):
        while True:
            rlist = list(self.notifiers.keys())
            if not rlist:
                break

            rlist += [sys.stdin]
            rlist, _, _ = select.select(rlist, [], [])
            if sys.stdin in rlist:
                rlist.remove(sys.stdin)
                if not self.get_keypress():
                    break
            try:
                for fd in rlist:
                    self.notifiers[fd].can_read(fd, self)

            # This exception occur when can_read have no more data available
            except ReadError:
                break

        curses.nocbreak()
        curses.echo()
        curses.endwin()



from PyQt5.QtWidgets import QApplication, QWidget, QMainWindow
from PyQt5.QtGui import QPainter, QColor, QPixmap
from PyQt5.QtGui import QIcon
from PyQt5.QtCore import Qt, QSocketNotifier, QSettings

from . import bagl
from .display import Display, DisplayArgs, FrameBuffer, COLORS, MODELS, ServerArgs
from .readerror import ReadError

BUTTON_LEFT = 1
BUTTON_RIGHT = 2
DEFAULT_WINDOW_X = 10
DEFAULT_WINDOW_Y = 10


class PaintWidget(QWidget):
    def __init__(self, parent, model, pixel_size, vnc=None):
        super().__init__(parent)
        self.fb = FrameBuffer(model)
        self.pixel_size = pixel_size
        self.mPixmap = QPixmap()
        self.vnc = vnc

    def paintEvent(self, event):
        if self.fb.pixels:
            pixmap = QPixmap(self.size() / self.pixel_size)
            pixmap.fill(Qt.white)
            painter = QPainter(pixmap)
            painter.drawPixmap(0, 0, self.mPixmap)
            self._redraw(painter)
            self.mPixmap = pixmap
            self.fb.pixels = {}

        qp = QPainter(self)
        copied_pixmap = self.mPixmap
        if self.pixel_size != 1:
            # Only call scaled if needed.
            copied_pixmap = self.mPixmap.scaled(
                self.mPixmap.width() * self.pixel_size,
                self.mPixmap.height() * self.pixel_size)
        qp.drawPixmap(0, 0, copied_pixmap)

    def update(self) -> bool:
        super().update()
        return self.fb.pixels != {}

    def _redraw(self, qp):
        for (x, y), color in self.fb.pixels.items():
            qp.setPen(QColor.fromRgb(color))
            qp.drawPoint(x, y)

        if self.vnc:
            self.vnc.redraw(self.fb.pixels)

        self.fb.screenshot_update_pixels()

    def draw_point(self, x, y, color):
        return self.fb.draw_point(x, y, color)

    def take_screenshot(self):
        return self.fb.take_screenshot()


class App(QMainWindow):
    def __init__(self, qt_app: QApplication, display: DisplayArgs, server: ServerArgs) -> None:
        super().__init__()

        self.setWindowTitle('Ledger %s Emulator' % MODELS[display.model].name)

        self.seph = server.seph
        self.width, self.height = MODELS[display.model].screen_size
        self.pixel_size = display.pixel_size
        self.box_position_x, self.box_position_y = MODELS[display.model].box_position
        box_size_x, box_size_y = MODELS[display.model].box_size

        # If the position of the window has been saved in the settings, restore
        # it.
        settings = QSettings("ledger", "speculos")
        # Take in account multiple screens and their geometry:
        current_screen_x = qt_app.primaryScreen().geometry().x()
        current_screen_y = qt_app.primaryScreen().geometry().y()
        if display.x is None:
            window_x = settings.value("window_x", current_screen_x + DEFAULT_WINDOW_X, int)
        else:
            window_x = display.x
        if display.y is None:
            window_y = settings.value("window_y", current_screen_y + DEFAULT_WINDOW_Y, int)
        else:
            window_y = display.y
        window_width = (self.width + box_size_x) * display.pixel_size
        window_height = (self.height + box_size_y) * display.pixel_size

        # Be sure Window is FULLY visible in one of the available screens:
        window_is_visible = False
        for screen in qt_app.screens():
            x1 = screen.geometry().x()
            y1 = screen.geometry().y()
            x2 = x1 + screen.geometry().width() - 1
            y2 = y1 + screen.geometry().height() - 1

            if window_x >= x1 and window_y >= y1 and (window_x + window_width - 1) <= x2 and \
               (window_y + window_height - 1) <= y2:
                window_is_visible = True
                break   # No need to check other screens

        # If the window is not FULLY visible, force default coordinates on current screen:
        if not window_is_visible:
            print("Window is NOT FULLY visible => using default coordinates in current screen.")
            window_x = current_screen_x + DEFAULT_WINDOW_X
            window_y = current_screen_y + DEFAULT_WINDOW_Y

        self.setGeometry(window_x, window_y, window_width, window_height)
        self.setFixedSize(window_width, window_height)

        flags = Qt.FramelessWindowHint
        if display.ontop:
            flags |= Qt.CustomizeWindowHint | Qt.WindowStaysOnTopHint
        self.setWindowFlags(flags)

        self.setAutoFillBackground(True)
        p = self.palette()
        p.setColor(self.backgroundRole(), QColor.fromRgb(COLORS[display.color]))
        self.setPalette(p)

        # Add paint widget and paint
        self.m = PaintWidget(self, display.model, display.pixel_size, server.vnc)
        self.m.move(self.box_position_x * display.pixel_size, self.box_position_y * display.pixel_size)
        self.m.resize(self.width * display.pixel_size, self.height * display.pixel_size)

        self.screen = Screen(self, display, server)

        self.setWindowIcon(QIcon('mcu/icon.png'))

        self.show()

    def screen_update(self) -> bool:
        return self.screen.screen_update()

    def keyPressEvent(self, event):
        self.screen._key_event(event, True)

    def keyReleaseEvent(self, event):
        self.screen._key_event(event, False)

    def _get_x_y(self):
        x = self.mouse_offset.x() // self.pixel_size - (self.box_position_x + 1)
        y = self.mouse_offset.y() // self.pixel_size - (self.box_position_y + 1)
        return x, y

    def mousePressEvent(self, event):
        '''Get the mouse location.'''

        self.mouse_offset = event.pos()

        x, y = self._get_x_y()
        if x >= 0 and x < self.width and y >= 0 and y < self.height:
            self.seph.handle_finger(x, y, True)
        QApplication.setOverrideCursor(Qt.DragMoveCursor)

    def mouseReleaseEvent(self, event):
        x, y = self._get_x_y()
        if x >= 0 and x < self.width and y >= 0 and y < self.height:
            self.seph.handle_finger(x, y, False)
        QApplication.restoreOverrideCursor()

    def mouseMoveEvent(self, event):
        '''Move the window.'''

        x = event.globalX()
        y = event.globalY()
        x_w = self.mouse_offset.x()
        y_w = self.mouse_offset.y()
        self.move(x - x_w, y - y_w)

    def closeEvent(self, event):
        '''
        Called when the window is closed. We save the current window position to
        the settings file in order to restore it upon next speculos execution.
        '''
        settings = QSettings("ledger", "speculos")
        settings.setValue("window_x", self.pos().x())
        settings.setValue("window_y", self.pos().y())


class Screen(Display):
    def __init__(self, app: App, display: DisplayArgs, server: ServerArgs) -> None:
        self.app = app
        super().__init__(display, server)
        self._init_notifiers(server)
        self.bagl = bagl.Bagl(app.m, MODELS[display.model].screen_size)
        self.seph = server.seph

    def klass_can_read(self, klass, s):
        try:
            klass.can_read(s, self)

        # This exception occur when can_read have no more data available
        except ReadError:
            self.app.close()

    def add_notifier(self, klass):
        n = QSocketNotifier(klass.s.fileno(), QSocketNotifier.Read, self.app)
        n.activated.connect(lambda s: self.klass_can_read(klass, s))

        assert klass.s.fileno() not in self.notifiers
        self.notifiers[klass.s.fileno()] = n

    def enable_notifier(self, fd, enabled=True):
        n = self.notifiers[fd]
        n.setEnabled(enabled)

    def remove_notifier(self, fd):
        # just in case
        self.enable_notifier(fd, False)

        n = self.notifiers.pop(fd)
        n.disconnect()

    def _key_event(self, event, pressed):
        key = event.key()
        if key in [Qt.Key_Left, Qt.Key_Right]:
            buttons = {Qt.Key_Left: BUTTON_LEFT, Qt.Key_Right: BUTTON_RIGHT}
            # forward this event to seph
            self.seph.handle_button(buttons[key], pressed)
        elif key == Qt.Key_Down:
            self.seph.handle_button(BUTTON_LEFT, pressed)
            self.seph.handle_button(BUTTON_RIGHT, pressed)
        elif key == Qt.Key_Q and not pressed:
            self.app.close()

    def display_status(self, data):
        ret = self.bagl.display_status(data)
        if MODELS[self.model].name == 'blue':
            self.screen_update()    # Actually, this method doesn't work
        return ret

    def display_raw_status(self, data):
        self.bagl.display_raw_status(data)
        if MODELS[self.model].name == 'blue':
            self.screen_update()    # Actually, this method doesn't work

    def screen_update(self) -> bool:
        return self.bagl.refresh()


class QtScreen:
    def __init__(self, display: DisplayArgs, server: ServerArgs) -> None:
        self.app = QApplication([])
        self.app_widget = App(self.app, display, server)
        self.m = self.app_widget.m

    def run(self):
        self.app.exec_()
        self.app.quit()



"""
Forward screen events (text, x, y) to TCP clients.
"""

import json
import logging
import socketserver
import threading


class AutomationServer(socketserver.ThreadingMixIn, socketserver.TCPServer):
    daemon_threads = True
    allow_reuse_address = True

    def __init__(self, server_address, RequestHandlerClass, *args, **kwargs):
        socketserver.TCPServer.__init__(self, server_address, RequestHandlerClass)

        self.logger = logging.getLogger("automation")
        self.clients = []

    def add_client(self, client):
        self.clients.append(client)

    def remove_client(self, client):
        self.clients.remove(client)

    def broadcast(self, event):
        """Broadcast an event to each connected client."""

        self.logger.debug(f"broadcast {event} to {self.clients}")
        for client in self.clients:
            client.send_screen_event(event)


class AutomationClient(socketserver.BaseRequestHandler):
    def setup(self):
        self.logger = logging.getLogger("automation")
        self.condition = threading.Condition()
        self.events = []
        self.logger.debug(f"new client from {self.client_address}")
        self.server.add_client(self)

    def handle(self):
        while True:
            with self.condition:
                while len(self.events) == 0:
                    self.condition.wait()

            while self.events:
                event = self.events.pop(0)
                data = json.dumps(event).encode()
                try:
                    self.request.sendall(data + b"\n")
                except BrokenPipeError:
                    return

    def finish(self):
        self.logger.debug("connection closed with client")
        self.server.remove_client(self)

    def send_screen_event(self, event):
        self.events.append(event)
        with self.condition:
            self.condition.notify()



from collections import namedtuple

'''
 {BAGL_GLYPH_ICON_DOWN, 7, 4, 1, C_icon_down_colors, C_icon_down_bitmap},
unsigned int const C_icon_down_colors[] = {
  0x00000000,
  0x00ffffff,

};

unsigned char const C_icon_down_bitmap[] = {
  0x41, 0x11, 0x05, 0x01,
};

typedef struct {
  unsigned int         icon_id;
  unsigned int         width;
  unsigned int         height;
  unsigned int         bits_per_pixel;
  const unsigned int*  default_colors; // color array entry count is (1<<bits_per_pixel)
  const unsigned char* bitmap;
} bagl_glyph_array_entry_t;
'''

Glyph = namedtuple("Glyph", "icon_id width height bpp colors bitmap")

BAGL_GLYPH_ICON_CHECK = 6
BAGL_GLYPH_ICON_CROSS = 7
BAGL_GLYPH_ICON_CHECK_BADGE = 8
BAGL_GLYPH_ICON_LEFT = 9
BAGL_GLYPH_ICON_RIGHT = 10
BAGL_GLYPH_ICON_UP = 11
BAGL_GLYPH_ICON_DOWN = 12
BAGL_GLYPH_ICON_CROSS_BADGE = 14
BAGL_GLYPH_ICON_TRANSACTION_BADGE = 24
BAGL_GLYPH_ICON_EYE_BADGE = 27

GLYPHS = [
    Glyph(BAGL_GLYPH_ICON_CHECK, 8, 6, 1, [0x00000000, 0x00ffffff], [0x80, 0x40, 0x20, 0x11, 0x0a, 0x04]),
    Glyph(BAGL_GLYPH_ICON_CROSS, 7, 7, 1, [0x00000000, 0x00ffffff], [0x41, 0x11, 0x05, 0x41, 0x11, 0x05, 0x01]),
    Glyph(BAGL_GLYPH_ICON_CHECK_BADGE, 14, 14, 1, [0x00ffffff, 0x00000000],
          [0x1f, 0xfe, 0x01, 0x3e, 0x00, 0x07, 0x80, 0x01, 0x24, 0x80, 0x83, 0x70, 0x70, 0x0e, 0xf8, 0x41, 0x3c, 0x18,
           0x06, 0x0e, 0xc0, 0x07, 0xf8, 0x87, 0x0f]),
    Glyph(BAGL_GLYPH_ICON_LEFT, 4, 7, 1, [0x00000000, 0x00ffffff], [0x48, 0x12, 0x42, 0x08]),
    Glyph(BAGL_GLYPH_ICON_RIGHT, 4, 7, 1, [0x00000000, 0x00ffffff], [0x21, 0x84, 0x24, 0x01]),
    Glyph(BAGL_GLYPH_ICON_UP, 7, 4, 1, [0x00000000, 0x00ffffff], [0x08, 0x8a, 0x28, 0x08]),
    Glyph(BAGL_GLYPH_ICON_DOWN, 7, 4, 1, [0x00000000, 0x00ffffff], [0x41, 0x11, 0x05, 0x01]),
    Glyph(BAGL_GLYPH_ICON_CROSS_BADGE, 14, 14, 1, [0x00ffffff, 0x00000000],
          [0x1f, 0xfe, 0x01, 0x3e, 0x00, 0x47, 0x88, 0x39, 0x27, 0xfc, 0x00, 0x1e, 0x80, 0x07, 0xf0, 0x43, 0xce, 0x19,
           0x21, 0x0e, 0xc0, 0x07, 0xf8, 0x87, 0x0f]),
    Glyph(BAGL_GLYPH_ICON_TRANSACTION_BADGE, 14, 14, 1, [0x00ffffff, 0x00000000],
          [0x1f, 0xfe, 0x01, 0x3e, 0x10, 0x07, 0x8c, 0xf9, 0x27, 0xfe, 0x01, 0x32, 0xc0, 0x04, 0xf8, 0x47, 0xfe, 0x19,
           0x03, 0x8e, 0xc0, 0x07, 0xf8, 0x87, 0x0f]),
    Glyph(BAGL_GLYPH_ICON_EYE_BADGE, 14, 14, 1, [0x00ffffff, 0x00000000],
          [0x1f, 0xfe, 0x01, 0x3e, 0x00, 0x07, 0x80, 0xe1, 0x21, 0xfe, 0xc1, 0xf3, 0xf0, 0x3c, 0xf8, 0x47, 0x78, 0x18,
           0x00, 0x0e, 0xc0, 0x07, 0xf8, 0x87, 0x0f])
]


def get(icon_id):
    for glyph in GLYPHS:
        if glyph.icon_id == icon_id:
            return glyph
    return None



from dataclasses import dataclass
import json
import jsonschema
import logging
import os
import pkg_resources
import re


@dataclass
class TextEvent:
    text: str
    x: int
    y: int


class Automation:
    def __init__(self, document):
        self.logger = logging.getLogger("automation")
        self.variables = {}

        if document.startswith("file:"):
            path = document[5:]
            with open(path) as fp:
                self.json = json.load(fp)
        else:
            self.json = json.loads(document)
        self.validate()

    def validate(self):
        path = os.path.join("resources", "automation.schema")
        with pkg_resources.resource_stream(__name__, path) as fp:
            schema = json.load(fp)
        jsonschema.validate(instance=self.json, schema=schema)

    def set_bool(self, key, value):
        self.variables[key] = value

    def get_actions(self, text, x, y):
        self.logger.debug(f'getting actions for "{text}" ({x}, {y})')

        for rule in self.json["rules"]:
            if "text" in rule and rule["text"] != text:
                continue
            if "regexp" in rule and not re.match(rule["regexp"], text):
                continue
            if "x" in rule and rule["x"] != x:
                continue
            if "y" in rule and rule["y"] != y:
                continue
            if "conditions" in rule:
                condition = True
                for key, value in rule["conditions"]:
                    if self.variables.get(key, False) != value:
                        condition = False
                        break
                if not condition:
                    continue

            if "actions" not in rule:
                self.logger.warning(f'missing "actions" key for rule {rule}')
                continue

            return rule["actions"]

        return []



import select
from typing import Optional

from . import bagl
from .display import Display, DisplayArgs, FrameBuffer, Model, MODELS, ServerArgs
from .readerror import ReadError
from .vnc import VNC


class Headless(Display):
    def __init__(self, display: DisplayArgs, server: ServerArgs) -> None:
        super().__init__(display, server)
        self._init_notifiers(server)

        self.m = HeadlessPaintWidget(self, self.model, server.vnc)
        self.bagl = bagl.Bagl(self.m, MODELS[self.model].screen_size)

    def display_status(self, data):
        ret = self.bagl.display_status(data)
        if MODELS[self.model].name == 'blue':
            self.screen_update()    # Actually, this method doesn't work
        return ret

    def display_raw_status(self, data):
        self.bagl.display_raw_status(data)
        if MODELS[self.model].name == 'blue':
            self.screen_update()    # Actually, this method doesn't work

    def screen_update(self) -> bool:
        return self.bagl.refresh()

    def run(self):
        while True:
            rlist = self.notifiers.keys()
            if not rlist:
                break

            rlist, _, _ = select.select(rlist, [], [])
            try:
                for fd in rlist:
                    self.notifiers[fd].can_read(fd, self)

            # This exception occur when can_read have no more data available
            except ReadError:
                break


class HeadlessPaintWidget(FrameBuffer):
    def __init__(self, parent: Headless, model: Model, vnc: Optional[VNC] = None) -> None:
        super().__init__(model)
        self.vnc = vnc

    def update(self):
        if self.pixels:
            self._redraw()
            self.pixels = {}
            return True
        return False

    def _redraw(self):
        if self.vnc:
            self.vnc.redraw(self.pixels)

        self.screenshot_update_pixels()



'''
TCP server to emulate buttons:

- L/R: press left/right button
- l/r: release left/right button

Usage example:

    $ echo -n RrRrRr | nc -nv 127.0.0.1 1235
    $ echo -n LRlr | nc -nv 127.0.0.1 1235
'''

import logging
import socket
import time


class FakeButtonClient:
    actions = {
        'L': (1, True),
        'l': (1, False),
        'R': (2, True),
        'r': (2, False),
    }

    def __init__(self, s):
        self.s = s
        self.logger = logging.getLogger("button")

    def _close(self, screen):
        screen.remove_notifier(self.s.fileno())
        self.logger.debug("connection closed with fake button client")

    def can_read(self, s, screen):
        packet = self.s.recv(1)
        if packet == b'':
            self._close(screen)
            return

        for c in packet:
            c = chr(c)
            if c in self.actions.keys():
                key, pressed = self.actions[c]
                self.logger.debug(f"button {key} release: {pressed}")
                screen.seph.handle_button(key, pressed)
                time.sleep(0.1)
            else:
                self.logger.debug(f"ignoring byte {c!r}")


class FakeButton:
    def __init__(self, port):
        self.s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.s.bind(('0.0.0.0', port))  # lgtm [py/bind-socket-all-network-interfaces]
        self.s.listen(5)
        self.logger = logging.getLogger("button")

    def can_read(self, s, screen):
        c, addr = self.s.accept()
        self.logger.debug(f"new client from {addr}")
        client = FakeButtonClient(c)
        screen.add_notifier(client)



from flask import Flask, request, stream_with_context, Response
from flask_restful import inputs, reqparse, Resource, Api
import json
import jsonschema
import io
import os.path
from PIL import Image
import pkg_resources
import socket
import threading
import time
from typing import Generator

from ..mcu import automation as mcu_automation
from ..mcu.readerror import ReadError


static_folder = pkg_resources.resource_filename(__name__, "/static")

app = Flask(__name__, static_url_path="", static_folder=static_folder)
app.env = "development"
api = Api(app)
automation_events = []


class Events:
    """This used to be the 'Automation Server'."""

    def __init__(self):
        self.clients = []
        self.events = []
        self.condition = threading.Condition()

    def add_client(self, client):
        app.logger.debug("events: new client")
        self.clients.append(client)

    def remove_client(self, client):
        app.logger.debug("events: client exited")
        self.clients.remove(client)

    def broadcast(self, event):
        app.logger.debug(f"events: broadcasting {event} to ({len(self.clients)}) client(s)")
        self.events.append(event)
        for client in self.clients:
            client.send_screen_event(event)
        with self.condition:
            self.condition.notify_all()


events = Events()


class ApiRunner:
    """Run the Speculos API server, with a notification when it stops"""
    def __init__(self, api_port: int) -> None:
        # self.s is used by Screen.add_notifier. Closing self._notify_exit
        # signals it that the API is no longer running.
        self.s, self._notify_exit = socket.socketpair()
        self.api_port = api_port

    def can_read(self, s: int, screen) -> None:
        assert s == self.s.fileno()
        # Being able to read from the socket only happens when the API server exited.
        raise ReadError("API server exited")

    def _run(self) -> None:
        try:
            # threaded must be set to allow serving requests along events streaming
            app.run(host="0.0.0.0", port=self.api_port, threaded=True, use_reloader=False)
        except Exception as exc:
            app.logger.error("an exception occurred in the Flask API server: %s", exc)
            raise
        finally:
            self._notify_exit.close()

    def start_server_thread(self, screen_, seph_) -> None:
        global screen, seph
        screen, seph = screen_, seph_
        seph.apdu_callbacks.append(apdu.seph_apdu_callback)
        api_thread = threading.Thread(target=self._run, name="API-server", daemon=True)
        api_thread.start()


def load_json_schema(filename):
    path = os.path.join("resources", filename)
    with pkg_resources.resource_stream(__name__, path) as fp:
        schema = json.load(fp)
    return schema


class Automation(Resource):
    def post(self):
        document = request.get_data()

        try:
            document = document.decode("ascii")
        except UnicodeDecodeError:
            return "invalid encoding", 400

        if document.startswith("file:"):
            return "invalid document", 400

        try:
            rules = mcu_automation.Automation(document)
        except json.decoder.JSONDecodeError:
            return "invalid document", 400
        except jsonschema.exceptions.ValidationError:
            return "invalid document", 400

        seph.automation = rules

        return {}, 200


class Button(Resource):
    schema = load_json_schema("button.schema")

    def post(self):
        args = request.get_json(force=True)
        try:
            jsonschema.validate(instance=args, schema=self.schema)
        except jsonschema.exceptions.ValidationError as e:
            return {"error": f"{e}"}, 400

        button = request.base_url.split("/")[-1]
        buttons = {"left": [1], "right": [2], "both": [1, 2]}
        action = args["action"]
        actions = {"press": [True], "release": [False], "press-and-release": [True, False]}
        delay = args.get("delay", 0.1)

        for a in actions[action]:
            for b in buttons[button]:
                seph.handle_button(b, a)
            if action == "press-and-release":
                time.sleep(delay)

        return {}, 200


class EventClient:
    def __init__(self):
        self.events = []

    def generate(self):
        try:
            # force headers to be sent
            yield b""

            while True:
                with events.condition:
                    events.condition.wait(1)

                while self.events:
                    event = self.events.pop(0)
                    data = json.dumps(event).encode()
                    yield data + b"\n"
        finally:
            events.remove_client(self)

    def send_screen_event(self, event):
        self.events.append(event)


class Events(Resource):
    def __init__(self):
        self.parser = reqparse.RequestParser()
        self.parser.add_argument("stream", type=inputs.boolean, default=False)
        super(Events, self).__init__()

    def get(self):
        args = self.parser.parse_args()
        if args.stream:
            client = EventClient()
            events.add_client(client)
            return Response(stream_with_context(client.generate()), content_type="text/event-stream")
        else:
            return {"events": events.events}, 200

    def delete(self):
        events.events.clear()
        return {}, 200


class Finger(Resource):
    schema = load_json_schema("finger.schema")

    def post(self):
        args = request.get_json(force=True)
        try:
            jsonschema.validate(instance=args, schema=self.schema)
        except jsonschema.exceptions.ValidationError as e:
            return {"error": f"{e}"}, 400

        action = args["action"]
        actions = {"press": [True], "release": [False], "press-and-release": [True, False]}
        delay = args.get("delay", 0.1)

        for a in actions[action]:
            seph.handle_finger(args["x"], args["y"], a)
            if action == "press-and-release":
                time.sleep(delay)

        return {}, 200


class APDU:
    def __init__(self):
        # We want to be notified when APDU response is transmitted from the SE
        self.endpoint_lock = threading.Lock()
        self.response_condition = threading.Condition()

    def exchange(self, data: bytes) -> Generator[bytes, None, None]:
        # force headers to be sent
        yield b""

        with self.endpoint_lock:  # Lock for a command/response for one client
            with self.response_condition:
                self.response = None
            seph.to_app(data)
            with self.response_condition:
                while self.response is None:
                    self.response_condition.wait()
            yield json.dumps({"data": self.response.hex()}).encode()

    def seph_apdu_callback(self, data: bytes):
        """
        Called by seph when data is transmitted by the SE. That data should
        be the response to a prior APDU request
        """
        with self.response_condition:
            self.response = data
            self.response_condition.notify()


apdu = APDU()


class APDU(Resource):
    schema = load_json_schema("apdu.schema")

    def post(self):
        args = request.get_json(force=True)
        try:
            jsonschema.validate(instance=args, schema=self.schema)
        except jsonschema.exceptions.ValidationError as e:
            return {"error": f"{e}"}, 400

        data = bytes.fromhex(args.get("data"))
        return Response(stream_with_context(apdu.exchange(data)), content_type="application/json")


class Screenshot(Resource):
    def get(self):
        screen_size, data = screen.m.take_screenshot()
        image = Image.frombytes("RGB", screen_size, data)
        iobytes = io.BytesIO()
        image.save(iobytes, format="PNG")
        return Response(iobytes.getvalue(), mimetype="image/png")


class Swagger(Resource):
    def get(self):
        return app.send_static_file("swagger/index.html")


class WebInterface(Resource):
    def get(self):
        return app.send_static_file("index.html")


api.add_resource(APDU, "/apdu")
api.add_resource(Automation, "/automation")
api.add_resource(Button, "/button/left", "/button/right", "/button/both")
api.add_resource(Events, "/events")
api.add_resource(Finger, "/finger")
api.add_resource(Screenshot, "/screenshot")
api.add_resource(Swagger, "/swagger/")
api.add_resource(WebInterface, "/")



'''
TCP server to emulate finger actions on screen:

args are sent in csv format (x,y,pressed) and can be chained.

Usage example:

    $ echo -n 128,300,1 | nc -nv 127.0.0.1 1236
    $ echo -n 128,300,0,110,41,0 | nc -nv 127.0.0.1 1236
'''

import logging
import socket


class FakeFingerClient:
    def __init__(self, s):
        self.s = s
        self.logger = logging.getLogger("finger")

    def _close(self, screen):
        screen.remove_notifier(self.s.fileno())
        self.logger.debug("connection closed with fake button client")

    def can_read(self, s, screen):
        packet = self.s.recv(100)
        if packet == b'':
            self._close(screen)
            return

        actions = packet.decode("ascii").split(',')
        actions = [actions[i * 3:(i + 1) * 3] for i in range(len(actions) // 3)]

        for action in actions:
            x = int(action[0])
            y = int(action[1])
            pressed = int(action[2])
            self.logger.debug(f"touch event on ({x},{y}) coordinates, {'pressed' if pressed else 'release'}")
            screen.seph.handle_finger(x, y, pressed)


class FakeFinger:
    def __init__(self, port):
        self.s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.s.bind(('0.0.0.0', port))  # lgtm [py/bind-socket-all-network-interfaces]
        self.s.listen(5)
        self.logger = logging.getLogger("finger")

    def can_read(self, s, screen):
        c, addr = self.s.accept()
        self.logger.debug(f"new client from {addr}")
        client = FakeFingerClient(c)
        screen.add_notifier(client)



#!/usr/bin/env python3

'''
Emulate the target app along the SE Proxy Hal server.
'''

import argparse
import binascii
import ctypes
from elftools.elf.elffile import ELFFile
import logging
from mnemonic import mnemonic
import os
import re
import signal
import socket
import sys
import threading

import pkg_resources

from .api import api
from .mcu import apdu as apdu_server
from .mcu import automation
from .mcu import display
from .mcu import seproxyhal
from .mcu.automation_server import AutomationClient, AutomationServer
from .mcu.button_tcp import FakeButton
from .mcu.finger_tcp import FakeFinger
from .mcu.vnc import VNC

DEFAULT_SEED = ('glory promote mansion idle axis finger extra february uncover one trip resource lawn turtle enact '
                'monster seven myth punch hobby comfort wild raise skin')

logger = logging.getLogger("speculos")

launcher_path = pkg_resources.resource_filename(__name__, "/resources/launcher")


def set_pdeath(sig):
    '''Set the parent death signal of the calling process.'''

    PR_SET_PDEATHSIG = 1
    libc = ctypes.cdll.LoadLibrary('libc.so.6')
    libc.prctl(PR_SET_PDEATHSIG, sig)


def get_elf_infos(app_path):
    with open(app_path, 'rb') as fp:
        elf = ELFFile(fp)
        text = elf.get_section_by_name('.text')
        symtab = elf.get_section_by_name('.symtab')
        bss = elf.get_section_by_name('.bss')
        sh_offset = text['sh_offset']
        sh_size = text['sh_size']
        stack = bss['sh_addr']
        sym_estack = symtab.get_symbol_by_name('_estack')
        if sym_estack is None:
            sym_estack = symtab.get_symbol_by_name('END_STACK')
        if sym_estack is None:
            logger.error('failed to find _estack/END_STACK symbol')
            sys.exit(1)
        estack = sym_estack[0]['st_value']
        supp_ram = elf.get_section_by_name('.rfbss')
        ram_addr, ram_size = (supp_ram['sh_addr'], supp_ram['sh_size']) if supp_ram is not None else (0, 0)
    stack_size = estack - stack
    return sh_offset, sh_size, stack, stack_size, ram_addr, ram_size


def run_qemu(s1: socket.socket, s2: socket.socket, args: argparse.Namespace) -> None:
    argv = ['qemu-arm-static']

    if args.debug:
        argv += ['-g', '1234', '-singlestep']

    argv += [launcher_path]

    if args.trace:
        argv += ['-t']

    if args.model is not None:
        argv += ['-m', args.model]

    argv += ['-k', str(args.sdk)]

    # load cxlib only if available for the specified sdk
    cxlib = pkg_resources.resource_filename(__name__, f"/cxlib/cx-{args.sdk}.elf")
    if os.path.exists(cxlib):
        argv += ['-c', cxlib]

    extra_ram = ''
    app_path = getattr(args, 'app.elf')
    for lib in [f'main:{app_path}'] + args.library:
        name, lib_path = lib.split(':')
        load_offset, load_size, stack, stack_size, ram_addr, ram_size = get_elf_infos(lib_path)
        # Since binaries loaded as libs could also declare extra RAM page(s), collect them all
        if (ram_addr, ram_size) != (0, 0):
            arg = f'{ram_addr:#x}:{ram_size:#x}'
            if extra_ram and arg != extra_ram:
                logger.error("different extra RAM pages for main app and/or libraries!")
                sys.exit(1)
            extra_ram = arg
        argv.append(f'{name}:{lib_path}:{load_offset:#x}:{load_size:#x}:{stack:#x}:{stack_size:#x}')

    if args.model == 'blue':
        if args.rampage:
            extra_ram = args.rampage
        if extra_ram:
            argv.extend(['-r', extra_ram])

    pid = os.fork()
    if pid != 0:
        return

    # ensure qemu is killed when this Python script exits
    set_pdeath(signal.SIGTERM)

    s2.close()

    # replace stdin with the socket
    os.dup2(s1.fileno(), sys.stdin.fileno())

    # handle both BIP39 mnemonics and hex seeds
    if args.seed.startswith("hex:"):
        seed = bytes.fromhex(args.seed[4:])
    else:
        seed = mnemonic.Mnemonic.to_seed(args.seed)

    os.environ['SPECULOS_SEED'] = binascii.hexlify(seed).decode('ascii')

    if args.deterministic_rng:
        os.environ['RNG_SEED'] = args.deterministic_rng

    logger.debug(f"executing qemu: {argv}")
    try:
        os.execvp(argv[0], argv)
    except FileNotFoundError:
        logger.error('failed to execute qemu: "%s" not found' % argv[0])
        sys.exit(1)
    sys.exit(0)


def setup_logging(args):
    logging.basicConfig(level=logging.INFO, format='%(asctime)s.%(msecs)03d:%(name)s: %(message)s', datefmt='%H:%M:%S')

    for arg in args.log_level:
        if ":" not in arg:
            logger.error(f"invalid --log argument {arg}")
            sys.exit(1)
        name, level = arg.split(":", 1)
        given_logger = logging.getLogger(name)
        try:
            given_logger.setLevel(level)
        except ValueError as e:
            logger.error(f"invalid --log argument: {e}")
            sys.exit(1)


def main():
    parser = argparse.ArgumentParser(description='Emulate Ledger Nano/Blue apps.')
    parser.add_argument('app.elf', type=str, help='application path')
    parser.add_argument('--automation', type=str, help='Load a JSON document automating actions (prefix with "file:" '
                                                       'to specify a path')
    parser.add_argument('--color', default='MATTE_BLACK', choices=list(display.COLORS.keys()), help='Nano color')
    parser.add_argument('-d', '--debug', action='store_true', help='Wait gdb connection to port 1234')
    parser.add_argument('--deterministic-rng', default="", help='Seed the rng with a given value to produce '
                                                                'deterministic randomness')
    parser.add_argument('-k', '--sdk', type=str, help='SDK version')
    parser.add_argument('-l', '--library', default=[], action='append', help='Additional library (eg. '
                        'Bitcoin:app/btc.elf) which can be called through os_lib_call')
    parser.add_argument('--log-level', default=[], action='append', help='Configure the logger levels (eg. usb:DEBUG), '
                                                                         'can be specified multiple times')
    parser.add_argument('-m', '--model', default='nanos', choices=list(display.MODELS.keys()))
    parser.add_argument('-r', '--rampage', help='Additional RAM page and size available to the app (eg. '
                                                '0x123000:0x100). Supersedes the internal probing for such page.')
    parser.add_argument('-s', '--seed', default=DEFAULT_SEED, help='BIP39 mnemonic or hex seed. Default to mnemonic: '
                                                                   'to use a hex seed, prefix it with "hex:"')
    parser.add_argument('-t', '--trace', action='store_true', help='Trace syscalls')
    parser.add_argument('-u', '--usb', default='hid', help='Configure the USB transport protocol, '
                                                           'either HID (default) or U2F')

    group = parser.add_argument_group('network arguments')
    group.add_argument('--apdu-port', default=9999, type=int, help='ApduServer TCP port')
    group.add_argument('--api-port', default=5000, type=int, help='Set the REST API TCP port (0 disables it)')
    group.add_argument('--automation-port', type=int, help='Forward text displayed on the screen to TCP clients')
    group.add_argument('--vnc-port', type=int, help='Start a VNC server on the specified port')
    group.add_argument('--vnc-password', type=str, help='VNC plain-text password (required for MacOS Screen Sharing)')
    group.add_argument('--button-port', type=int, help='Spawn a TCP server on the specified port to receive button '
                                                       'press (lLrR)')
    group.add_argument('--finger-port', type=int, help='Spawn a TCP server on the specified port to receive finger '
                                                       'touch (x,y,pressed)+')

    group = parser.add_argument_group('display arguments', 'These arguments might only apply to one of the display '
                                                           'method.')
    group.add_argument('--display', default='qt', choices=['headless', 'qt', 'text'])
    group.add_argument('--ontop', action='store_true', help='The window stays on top of all other windows')
    group.add_argument('--xy', action='store', help='Window position in "XxY" format (eg. "--xy 30x100").')
    group.add_argument('--keymap', action='store', help="Text UI keymap in the form of a string (e.g. 'was' => 'w' for "
                                                        "left button, 'a' right, 's' both). Default: arrow keys")
    group.add_argument('--progressive', action='store_true', help='Enable step-by-step rendering of graphical elements')
    group.add_argument('--zoom', help='Display pixel size.', type=int, choices=range(1, 11))

    args = parser.parse_args()
    args.model.lower()

    setup_logging(args)

    rendering = seproxyhal.RENDER_METHOD.FLUSHED
    if args.progressive:
        rendering = seproxyhal.RENDER_METHOD.PROGRESSIVE

    if args.rampage:
        if args.model != 'blue':
            logger.error("extra RAM page arguments -r (--rampage) require '-m blue'")
            sys.exit(1)
        if not re.match('(0x)?[0-9a-fA-F]+:(0x)?[0-9a-fA-F]+$', args.rampage):
            logger.error("invalid ram page argument")
            sys.exit(1)

    if args.display == 'text' and args.model not in ['nanos', 'nanox']:
        logger.error(f"unsupported model '{args.model}' with argument --display text")
        sys.exit(1)

    if args.ontop and args.display != 'qt':
        logger.error("--ontop can only be used with --display qt")
        sys.exit(1)

    if args.xy and args.display != 'qt':
        logger.error("--xy can only be used with --display qt")
        sys.exit(1)

    if args.zoom and args.display != 'qt':
        logger.error("--zoom can only be used with --display qt")
        sys.exit(1)

    if args.keymap and args.display != 'text':
        logger.error("--keymap can only be used with --display text")
        sys.exit(1)

    if args.vnc_password and not args.vnc_port:
        logger.error("--vnc-password can only be used with --vnc-port")
        sys.exit(1)

    if args.display == 'text':
        from .mcu.screen_text import TextScreen as Screen
    elif args.display == 'headless':
        from .mcu.headless import Headless as Screen
    else:
        from .mcu.screen import QtScreen as Screen

    if args.sdk is None:
        default_sdk = {
            "nanos": "2.0",
            "nanox": "1.2",
            "blue": "blue-2.2.5",
        }
        args.sdk = default_sdk.get(args.model)

    api_enabled = (args.api_port != 0)

    automation_path = None
    if args.automation:
        logger.warn("--automation is deprecated, please use the REST API instead")
        automation_path = automation.Automation(args.automation)

    automation_server = None
    if args.automation_port:
        logger.warn("--automation-port is deprecated, please use the REST API instead")
        if api_enabled:
            logger.warn("--automation-port is incompatible with the the API server, disabling it")
            api_enabled = False
        automation_server = AutomationServer(("0.0.0.0", args.automation_port), AutomationClient)
        automation_thread = threading.Thread(target=automation_server.serve_forever, daemon=True)
        automation_thread.start()

    if api_enabled:
        automation_server = api.events

    s1, s2 = socket.socketpair()

    run_qemu(s1, s2, args)
    s1.close()

    apdu = apdu_server.ApduServer(host="0.0.0.0", port=args.apdu_port)
    seph = seproxyhal.SeProxyHal(s2, automation=automation_path, automation_server=automation_server,
                                 transport=args.usb)

    button = None
    if args.button_port:
        logger.warn("--button-port is deprecated, please use the REST API instead")
        button = FakeButton(args.button_port)

    finger = None
    if args.finger_port:
        logger.warn("--finger-port is deprecated, please use the REST API instead")
        finger = FakeFinger(args.finger_port)

    vnc = None
    if args.vnc_port:
        screen_size = display.MODELS[args.model].screen_size
        vnc = VNC(args.vnc_port, screen_size, args.vnc_password)

    zoom = args.zoom
    if zoom is None:
        default_zoom = {
            "nanos": 2,
            "nanox": 2,
            "blue": 1,
        }
        zoom = default_zoom.get(args.model)

    x, y = None, None
    if args.xy:
        x, y = (int(i) for i in args.xy.split('x'))

    apirun = None
    if api_enabled:
        apirun = api.ApiRunner(args.api_port)

    display_args = display.DisplayArgs(args.color, args.model, args.ontop, rendering, args.keymap, zoom, x, y)
    server_args = display.ServerArgs(apdu, apirun, button, finger, seph, vnc)
    screen = Screen(display_args, server_args)

    if api_enabled:
        apirun.start_server_thread(screen, seph)

    screen.run()

    s2.close()


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# coding: utf8


class ReadError(Exception):
    """
    Raised when can_read method in SeProxyHal class have no more data available.
    """
    pass


class WriteError(Exception):
    """
    Raised when can_read method in SeProxyHal class have no more data available.
    """
    pass



'''
Forward packets between an external application and the emulated device.

Internally, it makes use of a TCP socket server to allow these 2 components to
communicate.
'''

import errno
import logging
import socket


class ApduServer:
    def __init__(self, host='127.0.0.1', port=9999, hid=False):
        self.s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.s.bind((host, port))  # lgtm [py/bind-socket-all-network-interfaces]
        self.s.listen()

        self.client = None

    def can_read(self, s, screen):
        assert self.s.fileno() == s

        c, _ = self.s.accept()
        self.client = ApduClient(c)
        screen.add_notifier(self.client)

    def forward_to_client(self, packet):
        if self.client is not None:
            self.client.forward_to_client(packet)


class ApduClient:
    def __init__(self, s):
        self.s = s
        self.logger = logging.getLogger("apdu")

    def _recvall(self, size):
        data = b''
        while size > 0:
            try:
                tmp = self.s.recv(size)
            except ConnectionResetError:
                tmp = b''
            if len(tmp) == 0:
                self.logger.debug("connection with client closed")
                return None
            data += tmp
            size -= len(tmp)
        return data

    def recv_packet(self):
        data = self._recvall(4)
        if data is None:
            return None

        size = int.from_bytes(data, byteorder='big')
        packet = self._recvall(size)
        if packet is None:
            return None

        return packet

    def can_read(self, s, screen):
        '''Forward APDU packet to the app'''

        assert self.s.fileno() == s

        packet = self.recv_packet()
        if packet is None:
            screen.remove_notifier(self.s.fileno())
            self.s.close()
            return

        self.logger.info("> {}".format(packet.hex()))
        screen.forward_to_app(packet)

    def forward_to_client(self, packet):
        '''Encode and forward APDU to the client.'''

        self.logger.info("< {}".format(packet.hex()))

        size = (len(packet) - 2) & 0xffffffff
        packet = size.to_bytes(4, 'big') + packet
        try:
            self.s.sendall(packet)
        except OSError as e:
            if e.errno == errno.EBADF:
                # the connection with the client was closed, ignore any error
                pass
            else:
                raise



'''
Forward framebuffer to a VNC server (vnc/vnc_server) and receive mouse events.

This VNC server can be tested with Remmina for instance (with color depth set to
true color: 24 bbp).
'''

import os
import logging
import subprocess
import sys


class VNC:
    def __init__(self, port, screen_size, password=None, verbose=False):
        self.logger = logging.getLogger("vnc")

        width, height = screen_size
        path = os.path.dirname(os.path.realpath(__file__))
        server = os.path.join(path, '../resources/vnc_server')
        cmd = [server]

        # custom options
        cmd += ['-s', f'{width}x{height}']
        if verbose:
            cmd += ['-v']

        # libvncserver options
        cmd += [
            '--',
            '-rfbport', f'{port}',
            '-rfbportv6', f'{port}',
        ]
        if password is not None:
            cmd += ['-passwd', password]

        self.p = subprocess.Popen(cmd, stdin=subprocess.PIPE, stdout=subprocess.PIPE)

        # required by Screen.add_notifier
        self.s = self.p.stdout

    def redraw(self, pixels):
        '''The framebuffer was updated, forward everything to the VNC server.'''

        # int.to_bytes() is super slow, hence the manual encoding
        buf = bytearray(len(pixels) * 9)
        i = 0
        for (x, y), color in pixels.items():
            buf[i + 0] = y & 0xff
            buf[i + 1] = (y >> 8) & 0xff
            buf[i + 2] = x & 0xff
            buf[i + 3] = (x >> 8) & 0xff
            buf[i + 4] = color & 0xff
            buf[i + 5] = (color >> 8) & 0xff
            buf[i + 6] = (color >> 16) & 0xff
            buf[i + 7] = (color >> 24) & 0xff
            buf[i + 8] = 0x0a
            i += 9

        self.p.stdin.write(buf)
        self.p.stdin.flush()

    def can_read(self, s, screen):
        '''Process a new keyboard or mouse event message from the VNC server.'''

        assert s == self.s.fileno()
        assert s == self.p.stdout.fileno()

        data = b''
        while len(data) != 6:
            tmp = self.p.stdout.read(6 - len(data))
            if not tmp:
                self.logger.info("connection with vnc stdout closed")
                sys.exit(0)
            data += tmp

        if data[4] in [0x00, 0x01]:
            # mouse
            x = int.from_bytes(data[0:2], 'little')
            y = int.from_bytes(data[2:4], 'little')
            pressed = (data[4] != 0x00)
            screen.seph.handle_finger(x, y, pressed)
        elif data[4] in [0x10, 0x11]:
            # keyboard
            button = data[0]
            pressed = (data[4] == 0x11)
            screen.seph.handle_button(button, pressed)
        else:
            self.logger.error("invalid message from the VNC server")



from typing import List, Mapping
from dataclasses import dataclass
import functools
import string

from .automation import TextEvent
from . import bagl_font

BitMap = bytes
BitVector = str  # a string of '1' and '0'
Width = int  # width
CharWidth = Width
Char = str  # a single character string (chars are 1-char strings in Python)


__FONT_MAP = {}

DISPLAY_CHARS = string.ascii_letters + string.digits + string.punctuation


def cache_font(f):
    __font_char_cache = {}

    @functools.wraps(f)
    def wrapper(byte_string: bytes):
        if byte_string not in __font_char_cache:
            __font_char_cache[byte_string] = f(byte_string)
        return __font_char_cache[byte_string]
    return wrapper


@dataclass
class BitMapChar:
    char: bagl_font.FontCharacter
    bitmap: bytes


def split(bits: BitVector, n: Width) -> List[BitVector]:
    """
    Split a bit array (string of '1' and '0')
    into an arbitrary list of n-sized bit arrays.

    Useful if you want to display a character from a bitmap
    """
    output = []
    bit_list = bits[:]
    while bit_list:
        bit_list_to_append = bit_list[:n]
        # do padding
        # bit_list_to_append += '0' * (n - len(bit_list_to_append))
        output.append(bit_list_to_append)
        bit_list = bit_list[n:]
    return output


def split_bytes(bytes_elems: BitMap, n: CharWidth):
    """Useful to display a bitmap character"""
    bits = "".join("{0:08b}".format(x)[::-1] for x in bytes_elems)
    return split(bits, n)


def get_font(font_id: int) -> bagl_font.Font:
    font = bagl_font.get(font_id)
    if not font:
        raise ValueError(f"Could not get font with font_id={font_id}")
    return font


def get_char(font: bagl_font.Font, char: str) -> BitMapChar:
    if len(char) != 1 or char not in DISPLAY_CHARS:
        raise ValueError(f"could not print {char}")
    return get_font_map(font)[char]


def display_char(font: bagl_font.Font, char: str) -> None:
    char = get_char(font, char)
    print("\n".join(split_bytes(char.bitmap, font.bpp * char.char.char_width)))


def get_font_map(font: bagl_font.Font):
    if font.font_id not in __FONT_MAP:
        __FONT_MAP[font.font_id] = _get_font_map(font)
    return __FONT_MAP[font.font_id]


def _get_font_map(font: bagl_font.Font) -> Mapping[Char, BitMapChar]:
    font_map = {}
    for ord_char, font_char in zip(
        range(font.first_char, font.last_char), font.characters
    ):
        font_map[chr(ord_char)] = BitMapChar(
            font_char,
            bytes(
                font.bitmap[
                    font_char.bitmap_offset:(
                        font_char.bitmap_offset
                        + font_char.bitmap_byte_count
                    )
                ]
            ),
        )
    return font_map


@cache_font
def find_char_from_bitmap(bitmap: BitMap):
    """
     Find a character from a bitmap
    >>> font = get_font(4)
    >>> char = get_char(font, 'c')
    >>> find_char_from_bitmap(char.bitmap)
    'c'
    """
    all_values = []
    for font in bagl_font.FONTS:
        font_map = get_font_map(font)
        for character_value, bitmap_struct in font_map.items():
            if bitmap_struct.bitmap.startswith(bitmap):
                # sometimes (but not always) the bitmap being passed is shortened
                # by one '\x00' byte, not matching the exact bitmap
                # provided in the font. Hence the 'residual' computation
                residual_bytes: bytes = bitmap_struct.bitmap[len(bitmap):]
                if all(b == 0 for b in residual_bytes):
                    all_values.append(character_value)
        if all_values:
            char = max([x for x in all_values])
            if char == "\x80":
                char = " "
            return char


class NanoXOCR:
    def __init__(self):
        self.events = []

    def analyze_bitmap(self, data: bytes):
        if data[0] != 0:
            return

        x = int.from_bytes(data[1:3], byteorder="big", signed=True)
        y = int.from_bytes(data[3:5], byteorder="big", signed=True)
        bpp = int.from_bytes(data[9:10], byteorder="big")
        color_size = 4 * (1 << bpp)
        bitmap = data[10+color_size:]

        char = find_char_from_bitmap(bitmap)
        if char:
            if self.events and y <= self.events[-1].y:
                self.events[-1].text += char
            else:
                # create a new TextEvent if there are no events yet or if there is a new line
                self.events.append(TextEvent(char, x, y))

    def get_events(self) -> List[TextEvent]:
        events = self.events.copy()
        self.events = []
        return events



"""
Forward USB packets between the MCU and the SE using the (custom) SE Proxy HAL
protocol.
"""

from abc import ABC, abstractmethod
from construct import Int8ub, Int16ub, Int16ul, Struct
import binascii
import enum
import logging


class UsbReq(enum.IntEnum):
    RECIPIENT_DEVICE = 0x00
    SET_ADDRESS = 0x05
    SET_CONFIGURATION = 0x09


class SephUsbTag(enum.IntEnum):
    XFER_SETUP = 0x01
    XFER_IN = 0x02
    XFER_OUT = 0x04
    XFER_EVENT = 0x10
    PREPARE_DIR_IN = 0x20


class SephUsbConfig(enum.IntEnum):
    CONNECT = 0x01
    DISCONNECT = 0x02
    ADDR = 0x03
    ENDPOINTS = 0x04


class SephUsbPrepare(enum.IntEnum):
    SETUP = 0x10
    IN = 0x20
    OUT = 0x30
    STALL = 0x40
    UNSTALL = 0x80


class HidEndpoint(enum.IntEnum):
    OUT_ADDR = 0x00
    IN_ADDR = 0x80


class UsbDevState(enum.IntEnum):
    DISCONNECTED = 0
    DEFAULT = 1
    ADDRESSED = 2
    CONFIGURED = 3


class UsbInterface(enum.IntEnum):
    GENERIC = 0
    U2F = 1
    HID = 2


USB_SIZE = 0x40

usb_header = Struct(
    "endpoint" / Int8ub,
    "tag" / Int8ub,
    "length" / Int8ub,
)

usb_setup = Struct(
    "bmreq" / Int8ub,
    "breq" / Int8ub,
    "wValue" / Int16ul,
    "wIndex" / Int16ub,
    "wLength" / Int16ub,
)

hid_header = Struct(
    "channel" / Int16ub,
    "command" / Int8ub,
    "seq" / Int16ub,
    "length" / Int16ub,
)


class HidPacket:
    def __init__(self):
        self.reset(0)

    def reset(self, size):
        self.data = b''
        self.seq = 0
        self.remaining_size = size

    def append_data(self, data):
        self.seq += 1
        if len(data) <= self.remaining_size:
            self.data += data
            self.remaining_size -= len(data)
        else:
            self.data += data[:self.remaining_size]
            self.remaining_size = 0

    def complete(self):
        return self.remaining_size == 0


class Transport(ABC):
    def __init__(self, interface, send_xfer):
        self.interface = interface
        self.send_xfer = send_xfer

    @abstractmethod
    def xfer(self, data):
        pass

    @abstractmethod
    def build_xfer(self, data):
        pass

    @abstractmethod
    def prepare(self, data):
        pass

    def config(self, tag):
        pass

    @property
    def endpoint_in(self):
        return HidEndpoint.IN_ADDR | self.interface

    @property
    def endpoint_out(self):
        return HidEndpoint.OUT_ADDR | self.interface


class U2f(Transport):
    def __init__(self, send_xfer):
        super().__init__(UsbInterface.U2F, send_xfer)

    def build_xfer(self, tag, data):
        packet = usb_header.build(dict(endpoint=self.endpoint_out, tag=tag, length=len(data)))
        packet += data
        return packet

    def xfer(self, data):
        assert len(data) == USB_SIZE
        packet = self.build_xfer(SephUsbTag.XFER_OUT, data)
        self.send_xfer(packet)

    def prepare(self, data):
        assert len(data) == USB_SIZE
        packet = self.build_xfer(SephUsbTag.XFER_IN, b'')
        self.send_xfer(packet)
        return data


class Hid(Transport):
    USB_CHANNEL = 0x0101
    USB_COMMAND = 0x05

    def __init__(self, send_xfer):
        super().__init__(UsbInterface.HID, send_xfer)
        self.hid_packet = HidPacket()

    def _build_header(self, data, length, seq):
        header = hid_header.build(dict(channel=self.USB_CHANNEL, command=self.USB_COMMAND, seq=seq, length=length))
        if seq != 0:
            # strip hid_header.length
            header = header[:-2]
        return header

    def build_xfer(self, tag, data, seq=0, length=USB_SIZE):
        header = self._build_header(data, length, seq)
        size = len(header) + len(data)

        packet = usb_header.build(dict(endpoint=self.endpoint_out, tag=tag, length=size))
        packet += header
        packet += data

        return packet

    def xfer(self, data):
        seq = 0
        offset = 0
        while offset < len(data):
            size = USB_SIZE - hid_header.sizeof() - 10
            if seq != 0:
                size += 2
            chunk = data[offset:offset+size]
            chunk = chunk.ljust(size, b'\x00')
            if seq == 0:
                length = len(data)
            else:
                length = len(chunk)

            packet = self.build_xfer(SephUsbTag.XFER_OUT, chunk, seq, length)
            self.send_xfer(packet)

            offset += len(chunk)
            seq += 1

    def config(self, tag):
        if tag == UsbDevState.DISCONNECTED:
            self.hid_packet.reset(0)

    def prepare(self, data):
        hid = hid_header.parse(data)
        assert hid.channel == self.USB_CHANNEL
        assert hid.command == self.USB_COMMAND

        if hid.seq == 0:
            self.hid_packet.reset(hid.length)
            chunk = data[hid_header.sizeof():]
        else:
            assert hid.seq == self.hid_packet.seq
            chunk = data[hid_header.sizeof() - 2:]

        self.hid_packet.append_data(chunk)
        packet = self.build_xfer(SephUsbTag.XFER_IN, b'', self.hid_packet.seq)
        self.send_xfer(packet)

        if self.hid_packet.complete():
            answer = self.hid_packet.data
            self.hid_packet.reset(0)
        else:
            answer = None

        return answer


class USB:
    def __init__(self, _queue_event_packet, transport='hid'):
        self._queue_event_packet = _queue_event_packet
        self.packets_to_send = []
        self.state = UsbDevState.DISCONNECTED

        if transport.lower() == 'hid':
            self.transport = Hid(self.send_xfer)
        elif transport.lower() == 'u2f':
            self.transport = U2f(self.send_xfer)
        else:
            raise ValueError(f"Unsupported USB transport {transport!r}")

        self.logger = logging.getLogger("usb")

    def send_xfer(self, packet):
        # don't send packets until the endpoint is configured
        if self.state != UsbDevState.CONFIGURED or len(self.packets_to_send) > 0:
            self.packets_to_send.append(packet)
            return

        self.logger.debug("[SEND_XFER] {}".format(binascii.hexlify(packet)))
        self._queue_event_packet(SephUsbTag.XFER_EVENT, packet)

    def _send_setup(self, breq, wValue):
        data = usb_header.build(dict(endpoint=self.transport.endpoint_out, tag=SephUsbTag.XFER_SETUP, length=0))
        data += usb_setup.build(dict(bmreq=UsbReq.RECIPIENT_DEVICE, breq=breq, wValue=wValue, wIndex=0, wLength=0))
        self.logger.debug("[SEND_SETUP] {}".format(binascii.hexlify(data)))
        self._queue_event_packet(SephUsbTag.XFER_EVENT, data)

    def _flush_packets(self):
        packets_to_send = self.packets_to_send
        self.packets_to_send = []
        for packet in packets_to_send:
            self.send_xfer(packet)

    def config(self, data):
        """Parse a config packet. If the endpoint address is set, configure it."""

        tag = SephUsbConfig(data[0])
        self.logger.debug("[CONFIG] {} {}".format(repr(tag), binascii.hexlify(data)))

        # The USB stack is shut down with USB_power(0) before being powered on.
        # Wait for the first CONNECT config message to ensure that USBD_Start()
        # has been called.
        if tag == SephUsbConfig.CONNECT:
            if self.state == UsbDevState.DISCONNECTED:
                self.state = UsbDevState.ADDRESSED
                self.logger.debug("set_address sent")
                self._send_setup(UsbReq.SET_ADDRESS, 1)

        elif tag == SephUsbConfig.DISCONNECT:
            self.state = UsbDevState.DISCONNECTED
            self.transport.config(tag)

        elif tag == SephUsbConfig.ADDR:
            if self.state == UsbDevState.ADDRESSED:
                self.state = UsbDevState.CONFIGURED
                self._send_setup(UsbReq.SET_CONFIGURATION, 1)
                self.logger.debug("configured")

        elif tag == SephUsbConfig.ENDPOINTS:
            # once the endpoint is configured, queued packets can be sent
            endpoint = data[2]
            if endpoint == self.transport.endpoint_out:
                self._flush_packets()

    def prepare(self, data):
        """Send or receive a packet chunk."""

        header = usb_header.parse(data[:3])
        answer = None
        tag = SephUsbPrepare(header.tag)
        self.logger.debug("[PREPARE] {} {} {}".format(repr(self.state), repr(tag), binascii.hexlify(data)))

        if tag == SephUsbPrepare.IN:
            if header.endpoint == self.transport.endpoint_in:
                assert header.length == USB_SIZE
                data = data[usb_header.sizeof():]
                answer = self.transport.prepare(data)

        return answer

    def xfer(self, data):
        self.transport.xfer(data)



from collections import namedtuple
from dataclasses import asdict
import logging
import sys
import time
import threading
from enum import IntEnum

from . import usb
from .nanox_ocr import NanoXOCR
from .readerror import ReadError, WriteError


class SephTag(IntEnum):
    BUTTON_PUSH_EVENT = 0x05
    FINGER_EVENT = 0x0c
    DISPLAY_PROCESSED_EVENT = 0x0d
    TICKER_EVENT = 0x0e
    CAPDU_EVENT = 0x16

    MCU = 0x31
    TAG_BLE_RADIO_POWER = 0x44
    SE_POWER_OFF = 0x46
    USB_CONFIG = 0x4f
    USB_EP_PREPARE = 0x50

    REQUEST_STATUS = 0x52
    RAPDU = 0x53
    PRINTC_STATUS = 0x5f

    GENERAL_STATUS = 0x60
    GENERAL_STATUS_LAST_COMMAND = 0x0000
    SCREEN_DISPLAY_STATUS = 0x65
    PRINTF_STATUS = 0x66
    SCREEN_DISPLAY_RAW_STATUS = 0x69

    FINGER_EVENT_TOUCH = 0x01
    FINGER_EVENT_RELEASE = 0x02


TICKER_DELAY = 0.1

RenderMethods = namedtuple('RenderMethods', 'PROGRESSIVE FLUSHED')
RENDER_METHOD = RenderMethods(0, 1)


def ticker(add_tick):
    """
    Function ran in a thread to trigger a tick at a periodic interval.

    There is no need to prevent clock drift.
    """

    logger = logging.getLogger("seproxyhal.ticker")
    flood = False

    while True:
        if not add_tick():
            if not flood:
                logger.warn("skipping ticker events (expected within a debugger)")
                flood = True
        else:
            flood = False
        time.sleep(TICKER_DELAY)


class PacketThread(threading.Thread):
    TICKER_PACKET = (SephTag.TICKER_EVENT, b'')

    def __init__(self, s, status_event, *args, **kwargs):
        super(PacketThread, self).__init__(name="packet", daemon=True)
        self.s = s
        self.queue_condition = threading.Condition()
        self.queue = []
        self.status_event = status_event
        self.logger = logging.getLogger("seproxyhal.packet")
        self.stop = False

    def _send_packet(self, tag, data=b''):
        """Send a packet to the app."""

        size = len(data).to_bytes(2, 'big')
        packet = tag.to_bytes(1, 'big') + size + data
        self.logger.debug("send {}" .format(packet.hex()))
        try:
            self.s.sendall(packet)
        except BrokenPipeError:
            self.stop = True
        except OSError:
            self.stop = True

    def queue_packet(self, tag, data=b'', priority=False):
        """
        Append a packet to the queue of packets to be sent.

        This function is meant to called by other threads.
        """

        if not priority:
            self.queue.append((tag, data))
        else:
            # Some status packets expect a specific event to be answered before
            # any other events. Put it at the beginning of the queue.
            self.queue.insert(0, (tag, data))

        # notify this thread that a new packet is available
        with self.queue_condition:
            self.queue_condition.notify()

    def add_tick(self):
        """Add a ticker event to the queue."""

        # Don't add too many ticker packets to the queue. For instance, the app
        # might be stuck if a breakpoint is hit within a debugger. It avoids
        # flooding the app on resume.
        if self.queue.count(self.TICKER_PACKET) > 10:
            return False

        self.queue_packet(SephTag.TICKER_EVENT)
        return True

    def run(self):
        while not self.stop:
            # wait for a status notification
            while not self.status_event.is_set():
                self.status_event.wait()
            self.status_event.clear()

            # wait for a packet to be available in the queue
            with self.queue_condition:
                while len(self.queue) == 0:
                    self.queue_condition.wait()

            tag, data = self.queue.pop(0)
            self._send_packet(tag, data)

        self.logger.debug("exiting")


class SeProxyHal:
    def __init__(self, s, automation=None, automation_server=None, transport='hid'):
        self.s = s
        self.logger = logging.getLogger("seproxyhal")
        self.printf_queue = ''
        self.automation = automation
        self.automation_server = automation_server

        self.status_event = threading.Event()
        self.packet_thread = PacketThread(self.s, self.status_event)
        self.packet_thread.start()

        self.ticker_thread = threading.Thread(name="ticker",
                                              target=ticker,
                                              args=(self.packet_thread.add_tick,),
                                              daemon=True)
        self.ticker_thread.start()
        self.usb = usb.USB(self.packet_thread.queue_packet, transport=transport)

        self.nanox_ocr = NanoXOCR()

        # A list of callback methods when an APDU response is received
        self.apdu_callbacks = []

    def _recvall(self, size):
        data = b''
        while size > 0:
            try:
                tmp = self.s.recv(size)
            except ConnectionResetError:
                tmp = b''

            if len(tmp) == 0:
                self.logger.debug("fd closed")
                return None
            data += tmp
            size -= len(tmp)
        return data

    def _send_packet(self, tag, data=b''):
        '''Send packet to the app.'''

        size = len(data).to_bytes(2, 'big')
        packet = tag.to_bytes(1, 'big') + size + data
        try:
            self.s.sendall(packet)
        except BrokenPipeError:
            raise WriteError("Broken pipe, failed to send data to the app")

    def apply_automation_helper(self, event):
        if self.automation_server:
            self.automation_server.broadcast(asdict(event))

        if self.automation:
            actions = self.automation.get_actions(event.text, event.x, event.y)
            for action in actions:
                self.logger.debug(f"applying automation {action}")
                key, args = action[0], action[1:]
                if key == "button":
                    self.handle_button(*args)
                elif key == "finger":
                    self.handle_finger(*args)
                elif key == "setbool":
                    self.automation.set_bool(*args)
                elif key == "exit":
                    self.s.close()
                    sys.exit(0)
                else:
                    assert False

    def apply_automation(self, events):
        for event in events:
            self.apply_automation_helper(event)

    def _close(self, s, screen):
        screen.remove_notifier(self.s.fileno())
        self.s.close()

    def can_read(self, s, screen):
        '''
        Handle packet sent by the app.

        This function is called thanks to a screen QSocketNotifier.
        '''

        assert s == self.s.fileno()

        data = self._recvall(3)
        if data is None:
            self._close(s, screen)
            raise ReadError("fd closed")

        tag = data[0]
        size = int.from_bytes(data[1:3], 'big')

        data = self._recvall(size)
        if data is None:
            self._close(s, screen)
            raise ReadError("fd closed")
        assert len(data) == size

        self.logger.debug(f"received (tag: {tag:#04x}, size: {size:#04x}): {data!r}")

        if tag & 0xf0 == SephTag.GENERAL_STATUS or tag == SephTag.PRINTC_STATUS:
            events = []
            if tag == SephTag.GENERAL_STATUS:
                if int.from_bytes(data[:2], 'big') == SephTag.GENERAL_STATUS_LAST_COMMAND:
                    if screen.screen_update():
                        if screen.model == "nanox":
                            events = self.nanox_ocr.get_events()

            elif tag == SephTag.SCREEN_DISPLAY_STATUS:
                self.logger.debug(f"DISPLAY_STATUS {data!r}")
                events = screen.display_status(data)
                self.packet_thread.queue_packet(SephTag.DISPLAY_PROCESSED_EVENT, priority=True)

            elif tag == SephTag.SCREEN_DISPLAY_RAW_STATUS:
                self.logger.debug("SephTag.SCREEN_DISPLAY_RAW_STATUS")
                screen.display_raw_status(data)
                if screen.model == "nanox":
                    self.nanox_ocr.analyze_bitmap(data)
                # https://github.com/LedgerHQ/nanos-secure-sdk/blob/1f2706941b68d897622f75407a868b60eb2be8d7/src/os_io_seproxyhal.c#L787
                #
                # io_seproxyhal_spi_recv() accepts any packet from the MCU after
                # having sent SCREEN_DISPLAY_RAW_STATUS. If some event (eg.
                # TICKER_EVENT) is replied before DISPLAY_PROCESSED_EVENT, it
                # will be silently ignored.
                #
                # A DISPLAY_PROCESSED_EVENT should be answered immediately,
                # hence priority=True.
                self.packet_thread.queue_packet(SephTag.DISPLAY_PROCESSED_EVENT, priority=True)
                if screen.rendering == RENDER_METHOD.PROGRESSIVE:
                    screen.screen_update()

            elif tag == SephTag.PRINTF_STATUS or tag == SephTag.PRINTC_STATUS:
                for b in [chr(b) for b in data]:
                    if b == '\n':
                        self.logger.info(f"printf: {self.printf_queue}")
                        self.printf_queue = ''
                    else:
                        self.printf_queue += b
                self.packet_thread.queue_packet(SephTag.DISPLAY_PROCESSED_EVENT, priority=True)
                if screen.rendering == RENDER_METHOD.PROGRESSIVE:
                    screen.screen_update()

            else:
                self.logger.error(f"unknown tag: {tag:#x}")
                sys.exit(0)

            # signal the sending thread that a status has been received
            self.status_event.set()

            # apply automation rules after having replied to the app
            if events:
                self.apply_automation(events)

        elif tag == SephTag.RAPDU:
            screen.forward_to_apdu_client(data)
            for c in self.apdu_callbacks:
                c(data)

        elif tag == SephTag.USB_CONFIG:
            self.usb.config(data)

        elif tag == SephTag.USB_EP_PREPARE:
            data = self.usb.prepare(data)
            if data:
                for c in self.apdu_callbacks:
                    c(data)
                screen.forward_to_apdu_client(data)

        elif tag == SephTag.MCU:
            pass

        elif tag == SephTag.TAG_BLE_RADIO_POWER:
            self.logger.warn("ignoring BLE tag")

        elif tag == SephTag.SE_POWER_OFF:
            self.logger.warn("received tag SE_POWER_OFF, exiting")
            self._close(s, screen)
            raise ReadError("SE_POWER_OFF")

        elif tag == SephTag.REQUEST_STATUS:
            # Ignore calls to io_seproxyhal_request_mcu_status()
            pass

        else:
            self.logger.error(f"unknown tag: {tag:#x}")
            sys.exit(0)

    def handle_button(self, button, pressed):
        '''Forward button press/release from the GUI to the app.'''

        if pressed:
            self.packet_thread.queue_packet(SephTag.BUTTON_PUSH_EVENT, (button << 1).to_bytes(1, 'big'))
        else:
            self.packet_thread.queue_packet(SephTag.BUTTON_PUSH_EVENT, (0 << 1).to_bytes(1, 'big'))

    def handle_finger(self, x, y, pressed):
        '''Forward finger press/release from the GUI to the app.'''

        if pressed:
            packet = SephTag.FINGER_EVENT_TOUCH.to_bytes(1, 'big')
        else:
            packet = SephTag.FINGER_EVENT_RELEASE.to_bytes(1, 'big')
        packet += x.to_bytes(2, 'big')
        packet += y.to_bytes(2, 'big')

        self.packet_thread.queue_packet(SephTag.FINGER_EVENT, packet)

    def to_app(self, packet):
        '''
        Forward raw APDU to the app.

        Packets can be forwarded directly to the SE thanks to
        SephTag.CAPDU_EVENT, but it doesn't work with messages are larger
        than G_io_seproxyhal_spi_buffer. Emulating a basic USB stack is more
        reliable, see issue #11.
        '''

        if packet.startswith(b'RAW!') and len(packet) > 4:
            tag, packet = packet[4], packet[5:]
            self.packet_thread.queue_packet(tag, packet)
        else:
            self.usb.xfer(packet)



#!/usr/bin/env python3
"""Install Speculos"""
import pathlib
from distutils.spawn import find_executable
from setuptools.command.build_py import build_py as _build_py
from setuptools import find_packages, setup
import sys
import tempfile


class BuildSpeculos(_build_py):
    """
    Extend "setup.py build_py" to build Speculos launcher and VNC server using cmake.

    This command requires some system dependencies (ARM compiler, libvncserver
    headers...) which are documented on https://speculos.ledger.com/installation/build.html

    distutils documentation about extending the build command:
    https://docs.python.org/3.8/distutils/extending.html#integrating-new-commands
    """

    def run(self):
        if not find_executable("cmake"):
            raise RuntimeError("cmake is not found and is required to build Speculos")

        if not self.dry_run:
            pathlib.Path(self.build_lib).mkdir(parents=True, exist_ok=True)
            with tempfile.TemporaryDirectory(prefix="build-", dir=self.build_lib) as build_dir:
                self.spawn(
                    [
                        "cmake",
                        "-H.",
                        "-B" + build_dir,
                        "-DCMAKE_BUILD_TYPE=Release",
                        "-DBUILD_TESTING=0",
                        "-DWITH_VNC=1",
                    ]
                )
                self.spawn(["cmake", "--build", build_dir])

        super().run()


setup(
    name="speculos",
    author="Ledger",
    author_email="hello@ledger.fr",
    version="0.1.0",
    url="https://github.com/LedgerHQ/speculos",
    python_requires=">=3.6.0",
    description="Ledger Blue and Nano S/X application emulator",
    long_description=pathlib.Path("README.md").read_text(),
    long_description_content_type="text/markdown",
    packages=find_packages(),
    install_requires=[
        "construct>=2.10.56,<3.0.0",
        "flask>=2.0.0,<3.0.0",
        "flask-restful>=0.3.8,<1.0",
        "jsonschema>=3.2.0,<4.0.0",
        "mnemonic>=0.19,<1.0",
        "pillow>=8.0.0,<9.0.0",
        "pyelftools>=0.27,<1.0",
        "pyqt5>=5.15.2,<6.0.0",
        "requests>=2.25.1,<3.0.0",
    ]
    + (["dataclasses>=0.8,<0.9"] if sys.version_info <= (3, 6) else []),
    setup_requires=["wheel"],
    entry_points={
        "console_scripts": [
            "speculos = speculos.main:main",
        ],
    },
    cmdclass={
        "build_py": BuildSpeculos,
    },
    include_package_data=True,
)
